from bustoken import TK_SITEF_ADD_PAYMENT
from msgbus import FM_PARAM, TK_SYS_ACK
from mw_helper import show_message_dialog
from sysactions import get_used_service, show_messagebox, close_asynch_dialog
from systools import sys_log_exception


def request_eft(pos_id, mb_context, model, is_credit, amount, _eft_events, is_mercadopago=False):
    order = model.find("CurrentOrder/Order")
    totaled_state = order.findall("StateHistory/State[@state='TOTALED']")[-1]
    last_totaled = totaled_state.get("timestamp").replace("-", "").replace(":", "")

    if is_credit:
        tender_type = 1
    else:
        tender_type = 122 if is_mercadopago else 2

    order_id = order.get("orderId")
    tender_seq_id = len(order.find("TenderHistory").findall("Tender"))
    fiscal_date = last_totaled[:8]
    fiscal_time = last_totaled[9:15]
    operator = model.find("Operator").get("id")
    eft_name = get_used_service(model, "eft")

    dialog_id = show_message_dialog(pos_id, "$INFORMATION", "$WAITING_SITEF_ADD_PAYMENT", "NOCONFIRM", True)
    try:
        component_name = eft_name or "Sitef%02d" % int(pos_id)
        evt_data = [eft_name, None, pos_id, order_id]
        _eft_events.append(evt_data)

        data = "%s;%s;%s;%s;%s;%s;%s;%s;%s" % (str(pos_id), str(order_id), str(operator), str(tender_type), str(amount),
                                               fiscal_date, fiscal_time, tender_seq_id, False)
        ret = mb_context.MB_EasySendMessage(component_name, TK_SITEF_ADD_PAYMENT, FM_PARAM, data)
        if ret.token != TK_SYS_ACK:
            return
        else:
            return ret.data
    except Exception as _: # noqa
        show_messagebox(pos_id, "$ERROR_PROCESSING_TEF", "$EFT", asynch=True, buttons="$OK")
        sys_log_exception("Error processing TEF")
    finally:
        close_asynch_dialog(pos_id, dialog_id)



from sysactions import get_model, get_posot

from application.model.order import Order


class VerifyAndRemoveBenefitsUnitOfWork(object):
    def __init__(self, pos_id, order):
        # type: (int, Order) -> None

        self.pos_id = pos_id
        self.model = get_model(self.pos_id)
        self.pos_ot = get_posot(self.model)
        self.order = order



# -*- coding: utf-8 -*-

from sysactions import action, show_confirmation, show_messagebox, StopAction
from bustoken import TK_REMOTE_ORDER_SEND_ORDER_TO_LOGISTIC
from msgbus import TK_SYS_ACK

from .. import mb_context


@action
def send_logistic_remote_order(pos_id, order_id=""):
    confirmation = show_confirmation(pos_id, "$LOGISTIC_SEND_CONFIRMATION")
    if not confirmation:
        raise StopAction()

    data = pos_id + "\0" + order_id
    logistic_send_order_token = TK_REMOTE_ORDER_SEND_ORDER_TO_LOGISTIC
    msg = mb_context.MB_EasySendMessage("RemoteOrder", logistic_send_order_token, data=data, timeout=10000000)
    if msg.token == TK_SYS_ACK:
        show_messagebox(pos_id, "$LOGISTIC_SENT")
        return

    show_messagebox(pos_id, "$LOGISTIC_SENT_ERROR")



from logging import getLogger

from messagebus import Message, DefaultToken, DataType
from messageprocessorutil import UuidMessageProcessor

from nperipherals_core.interactor.print_interactor import PrintInteractor

logger = getLogger(__name__)


class PrintProcessor(UuidMessageProcessor):
    def __init__(self, interactor, event_name=None):
        # type: (PrintInteractor, str) -> None

        super(PrintProcessor, self).__init__(event_name)

        self.interactor = interactor

    def parse_data(self, data, event_type):
        return data

    def call_business(self, obj):
        self.interactor.print_data(obj)

    def format_response(self, message_bus, message, event, input_obj, result):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_ACK, DataType.string, "ok"))
        logger.info("PrintProcessor ACK")

    def format_exception(self, message_bus, message, event, input_obj, exception):
        self._reply_nak(exception, message, message_bus)

    def format_parse_exception(self, message_bus, message, event, data, exception):
        self._reply_nak(exception, message, message_bus)

    @staticmethod
    def _reply_nak(exception, message, message_bus):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, repr(exception)))
        logger.error(repr(exception))



from helper import ExtendedEnum
from messagebus import TokenCreator, TokenPriority

MESSAGE_GROUP = "112"


class ListenedTokens(ExtendedEnum):
    TK_NDISCOUNT_ADD_BENEFIT = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "1")
    TK_NDISCOUNT_REMOVE = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "2")
    TK_NDISCOUNT_GET_BENEFITS = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "7")
    TK_NDISCOUNT_ON_ITEM_SALE = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "8")
    TK_NDISCOUNT_ON_ORDER_TOTAL = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "9")
    TK_NDISCOUNT_ON_ORDER_REOPEN = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "10")
    TK_NDISCOUNT_ON_TABLE_REOPEN = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "12")
    TK_NDISCOUNT_ON_TABLE_OPEN = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "13")



from interactor import BaseInteractor
from model.resync import ResyncRequest


class ResyncInteractor(BaseInteractor):
    def execute(self, resync_request):
        # type: (ResyncRequest) -> None
        events_to_dispatch = self.log_repository.fetch_resync_events(resync_request)
        self.event_dispatcher.dispatch_events(events_to_dispatch)



from mw_helper import ensure_iterable

from ._OrderCondition import OrderCondition
from production.model import ProductionOrder


class CustomPropertyCondition(OrderCondition):
    def __init__(self, properties):
        self.properties = ensure_iterable(properties)

    def evaluate(self, order):
        # type: (ProductionOrder) -> bool

        for prop in self.properties:
            if prop in order.properties:
                return True

        return False



# -*- coding: utf-8 -*-
import json
import re
from typing import Dict

from sysactions import show_any_dialog, show_keyboard, show_messagebox

from actions.config import get_customer_info_config
from datetime import datetime, timedelta

from actions.void.utils import get_void_reason_options


def get_void_reason(pos_id, selected_index=None):
    void_reason_ids = range(1, 13)

    void_reason_options = get_void_reason_options(pos_id)
    void_reason_to_show = void_reason_options[0:4]
    if not selected_index:
        selected_index = show_any_dialog(pos_id, "filteredListBox", "", "$SELECT_REASON", "|".join(void_reason_to_show),
                                         "$OK|$CANCEL", "", 600000, "NOFILTER", "", "", "", None, False)

    if selected_index is None:
        return None

    selected_index = int(selected_index)
    void_reason_id = void_reason_ids[int(selected_index)]
    void_reason_descr = void_reason_options[int(selected_index)]
    void_reason_dict = {'VOID_REASON_ID': void_reason_id, 'VOID_REASON_DESCR': void_reason_descr}
    return void_reason_dict


def update_custom_properties(customer_doc=None, customer_name=None, pre_sale=None):
    dict_sale = {}

    if customer_doc not in (None, ""):
        dict_sale.update({"CUSTOMER_DOC": customer_doc})
    if customer_name not in (None, ""):
        dict_sale.update({"CUSTOMER_NAME": customer_name})
    if pre_sale not in (None, ""):
        dict_sale.update({"PRE_VENDA": pre_sale})

    return dict_sale


def get_customer_doc_after_paid(pos_id, pos_ot):
    customer_info_config = get_customer_info_config(pos_id)

    if "afterPaid" not in customer_info_config['document']:
        return ""

    valid = False
    customer_doc = ""
    while not valid:
        customer_doc = show_keyboard(pos_id, "$ONLY_NUMBER", "$ENTER_CPF_CNPJ", mask="CPF_CNPJ", numpad=True)
        if customer_doc is None:
            return None

        if customer_doc == "":
            ret = show_messagebox(pos_id, "$CONFIRM_EMPTY_CPF_CNPJ", buttons="$YES|$NO")
            if ret == 1:
                continue
            break
        customer_doc = str(customer_doc)
        valid = validate_document(customer_doc)
        if not valid:
            show_messagebox(pos_id, "$INVALID_CPF_CNPJ")

    fill_additional_info(pos_ot, "CPF", customer_doc)
    return customer_doc


def validate_document(customer_doc):
    valid = False
    if len(customer_doc) <= 11:
        valid = validar_cpf(customer_doc)
    elif len(customer_doc) > 11:
        valid = validar_cnpj(customer_doc)
    return valid


def fill_additional_info(pos_ot, properties, value):
    if pos_ot.additionalInfo:
        pos_ot.additionalInfo += "|"
    else:
        pos_ot.additionalInfo = ""
    pos_ot.additionalInfo += "{0}={1}".format(properties, value)


def validar_cpf(cpf):
    cpf = ''.join(re.findall('\d', str(cpf)))

    if (not cpf) or (len(cpf) < 11):
        return False

    inteiros = map(int, cpf)
    novo = inteiros[:9]

    while len(novo) < 11:
        r = sum([(len(novo) + 1 - i) * v for i, v in enumerate(novo)]) % 11

        if r > 1:
            f = 11 - r
        else:
            f = 0
        novo.append(f)

    if novo == inteiros:
        return cpf

    return False


def validar_cnpj(cnpj):
    cnpj = ''.join(re.findall('\d', str(cnpj)))

    if (not cnpj) or (len(cnpj) < 14):
        return False

    inteiros = map(int, cnpj)
    novo = inteiros[:12]

    prod = [5, 4, 3, 2, 9, 8, 7, 6, 5, 4, 3, 2]
    while len(novo) < 14:
        r = sum([x * y for (x, y) in zip(novo, prod)]) % 11
        if r > 1:
            f = 11 - r
        else:
            f = 0
        novo.append(f)
        prod.insert(0, 6)

    if novo == inteiros:
        return cnpj

    return False


def json_serialize(dictionary):
    # type: (Dict) -> str
    try:
        return json.dumps(dictionary, default=lambda o: getattr(o, '__dict__', str(o)))
    except ValueError as _:
        return dictionary


def compare_objects(obj1, obj2):
    return object_to_json(obj1) == object_to_json(obj2)


def object_to_json(obj):
    return json.dumps(obj, sort_keys=True, default=lambda o: getattr(o, '__dict__', str(o)))


def get_seconds_to_renew_cache(seconds_to_renew_cache):
    seconds_to_renew = 60 * 5
    try:
        seconds_to_renew = int(seconds_to_renew_cache)
    except TypeError as _:
        pass
    return seconds_to_renew


def reached_seconds_to_renew_cache(last_cache_time, seconds_to_renew_cache):
    return not last_cache_time or datetime.now() > last_cache_time + timedelta(seconds=seconds_to_renew_cache)



import json
import logging

from xml.etree import cElementTree as eTree

from actions.sale_line.set_custom_properties import set_custom_properties
from bustoken import TK_NCOUPON_APPLY_COUPON
from mw_helper import show_keyboard
from sysactions import action, get_model, get_current_order, show_messagebox, get_posot
from msgbus import FM_PARAM, TK_SYS_ACK, MBTimeout, MBException

from actions import mb_context
from custom_sysactions import user_authorized_to_execute_action

logger = logging.getLogger("PosActions")


@action
@user_authorized_to_execute_action
def do_apply_coupons(pos_id, table_id=None, seat=''):
    coupon_code = show_keyboard(pos_id, "$ENTER_COUPON_CODE", defvalue="", title="", numpad=False)
    if coupon_code is None:
        return False

    if seat == '':
        apply_coupon(pos_id, table_id=table_id, coupon_code=coupon_code)
        return

    _apply_coupon_and_set_seat(coupon_code, pos_id, seat, table_id)


def apply_coupon(pos_id, table_id=None, coupon_code=None):
    model = get_model(pos_id)
    order = get_current_order(model)
    message = None
    try:
        coupon_code = coupon_code.upper()
        originator_pos_id = int(order.get("originatorId")[3:]) if order else None
        data = json.dumps(dict(posId=pos_id, originatorPosId=originator_pos_id, voucherId=coupon_code))
        try:
            msg = mb_context.MB_EasySendMessage("nCoupon", TK_NCOUPON_APPLY_COUPON, FM_PARAM, data)
            message = "$SUCCESS_NOT_TOTALED_COUPON_APPLIED"
            if msg.token != TK_SYS_ACK:
                message = "${}".format(msg.data)
                return False
            if msg.token == TK_SYS_ACK and order and order.get('state') == "TOTALED":
                message = "$SUCCESS_COUPON_APPLIED"
                if table_id is not None:
                    from table_actions import process_discount_in_table
                    process_discount_in_table(pos_id, order, table_id)
                return True

        except (MBTimeout, MBException) as _:
            show_messagebox(pos_id, "$ERROR_COUPON_APPLIED")
            logger.exception("Error sending message to nDiscount")
            return False
    except Exception as _:
        show_messagebox(pos_id, "$ERROR_COUPON_APPLIED")
        logger.exception("Error applying coupon")
        return False
    finally:
        if message:
            show_messagebox(pos_id, message)


def _apply_coupon_and_set_seat(coupon_code, pos_id, seat, table_id):
    # type: (str, str, str, str) -> None

    model = get_model(pos_id)
    pos_ot = get_posot(model)
    order_picture = eTree.XML(pos_ot.orderPicture(pos_id))
    order_id = order_picture.get("orderId")
    sale_lines = order_picture.findall(".//SaleLine")
    last_line_number = max([int(x.get("lineNumber")) for x in sale_lines]) if sale_lines else 0
    apply_coupon(pos_id, table_id=table_id, coupon_code=coupon_code)
    _set_custom_properties_for_added_sale_lines(pos_id, pos_ot, order_id, last_line_number, seat)


def _set_custom_properties_for_added_sale_lines(pos_id, pos_ot, order_id, last_line_number, seat):
    order_picture = eTree.XML(pos_ot.orderPicture(pos_id, order_id))
    sale_lines = order_picture.findall(".//SaleLine")
    last_line_number_after_apply_coupon = max([int(x.get("lineNumber")) for x in sale_lines]) if sale_lines else 0
    if last_line_number != last_line_number_after_apply_coupon:
        sale_type = None if order_picture.get("saleTypeDescr") == "EAT_IN" else order_picture.get("saleTypeDescr")

        for sale_line in sale_lines:
            if int(sale_line.get("level")) > 0 or int(sale_line.get("lineNumber")) <= int(last_line_number):
                continue

            set_custom_properties(pos_ot, sale_line, seat, sale_type)



from mw_helper import ensure_iterable
from production.box._OrderChangerProductionBox import OrderChangerProductionBox
from production.model import ProdStates, ProductionOrder


class CustomPropertyValueFilterBox(OrderChangerProductionBox):
    def __init__(self, name, sons, parameter_name, allowed_values, excluded_values=None, logger=None):
        super(CustomPropertyValueFilterBox, self).__init__(name, sons, logger)
        self.parameter_name = parameter_name
        self.allowed_values = [x.lower() for x in ensure_iterable(allowed_values or [])]
        self.excluded_values = [x.lower() for x in ensure_iterable(excluded_values or [])]

    def change_order(self, order):
        if self.should_filter_order(order):
            order.prod_state = ProdStates.INVALID
        return order

    def should_filter_order(self, order):
        # type: (ProductionOrder) -> bool
        self.debug("should filter order, parameter = '{}'".format(self.parameter_name))
        if self.parameter_name in order.properties:
            prop_value = order.properties[self.parameter_name].lower()
            self.debug("parameter value = '{}'".format(prop_value))
            self.debug("allowed = '{}', excluded = '{}'".format(self.allowed_values, self.excluded_values))
            if self.allowed_values and prop_value not in self.allowed_values:
                return True
            if self.excluded_values and prop_value in self.excluded_values:
                return True
        return False



import cfgtools

from util import find_key_value, find_key_values

from ._BumpBarCommand import BumpBarCommand


class BumpBar(object):
    def __init__(self, group):
        # type: (cfgtools.Group) -> BumpBar
        self.name = ''
        self.simulate_bump_bar = True
        self.commands = []

        self.parse_bump_bar(group)

    def parse_bump_bar(self, bump_bar_group):
        # type: (cfgtools.Group) -> ()
        self.name = bump_bar_group.name
        self.simulate_bump_bar = find_key_value(bump_bar_group, "SimulateBumpBar").lower() == 'true'
        command_group = next((x for x in bump_bar_group.groups if x.name == "Commands"), [])

        for command in command_group.keys:
            self.commands.append(BumpBarCommand(command.name, command.values[0].split("|")))




from _AuditEvent import AuditEvent
from _EventData import EventData
from _OperatorCashInOut import OperatorCashInOut
from _Command import Command
from _TransferType import TransferType
from _TransferOrderInfo import TransferOrderInfo



class OrderItem(object):
    def __init__(self, name, quantity, pcode, unit_price):
        # type: (unicode, float, int, float) -> None
        self.name = name
        self.quantity = quantity
        self.pcode = pcode
        self.unit_price = unit_price



from msgbus import TK_POS_BUSINESSEND, FM_PARAM, TK_SYS_ACK
from mw_helper import show_message_dialog, show_message_options_dialog
from sysactions import action, get_model, get_business_period, set_custom, print_report, translate_message, send_message
from systools import sys_log_info, sys_log_error, sys_log_exception
from bustoken import TK_POS_HAS_BINARY_UPDATE_TO_APPLY, TK_POS_UPDATER_PERFORM_BINARIES_UPDATE


from actions.util import list_opened_users
from actions.pos.check_table_status import has_tab_or_tables_opened, show_open_table_dialog
from actions.util.get_menu_user_logged import get_menu_manager_authenticated_user
from actions.models.mw_sys_errors import MwSysErrors
from datetime import datetime


@action
def close_day(pos_id):
    from msgbusboundary import MsgBusTableService as tableService

    error_dict = {}
    model = get_model(pos_id)
    period = get_business_period(model)
    is_table_service = model.find("WorkingMode").get("usrCtrlType") == "TS"

    pos_list = [pos_id]
    if is_table_service:
        pos_list = tableService().get_exclusive_pos_list(user_control_type="TS")

    if is_table_service and has_tab_or_tables_opened(period, pos_id):
        show_open_table_dialog(pos_id)
        return False

    opened_users = list_opened_users(pos_id)
    _fill_pos_list_with_can_close_day_pos(error_dict, period, pos_list, opened_users)
    error_messages = _parse_error_messages(model, error_dict)

    if len(error_messages) > 0:
        show_message_dialog(pos_id, message="$CANNOT_CLOSE_DAYS|{}".format("\\".join(error_messages)))
        return False

    if len(pos_list) == 0:
        show_message_dialog(pos_id, message="$HAS_NO_POS_TO_CLOSE")
        return False

    close_date = datetime.strptime(period, '%Y%m%d').strftime('%d/%m/%Y')
    pos_list = [int(x) for x in pos_list]
    confirmation_message = "$CONFIRM_CLOSE_DAY|{0}|{1}".format(close_date, "POS {}".format(pos_list))
    resp = show_message_options_dialog(pos_id, "$OK|$CANCEL", "$INFORMATION", confirmation_message)
    if resp is None or resp == 1:
        return False

    try:
        sys_log_info("Closing business day for POS's: {}".format(pos_list))
        error_dict = {}
        succeed_pos_id = []

        for pos_no in pos_list:
            _perform_close_day(pos_no, succeed_pos_id, error_dict)
            set_custom(pos_no, 'POS_MUST_BE_BLOCKED', False)

        if len(error_dict) > 0:
            error_messages = _parse_error_messages(model, error_dict)
            show_message_dialog(pos_id, "$INFORMATION", "$ERROR_CLOSING_DAYS|{}".format("\\".join(error_messages)))

        if len(succeed_pos_id) > 0:
            report_pos_id = "|".join(map(str, succeed_pos_id))
            menu_manager_user = get_menu_manager_authenticated_user()
            print_report(pos_id, model, True, "cash", pos_id, period, "0", "true", report_pos_id, "end_of_day", "",
                         menu_manager_user)

            pending_updates = _has_pending_pos_update(pos_id)
            if pending_updates == "HAS_PENDING_UPDATE":
                confirmation_message = "$HAS_PENDING_UPDATES"
                resp = show_message_options_dialog(pos_id, "$OK|$CANCEL", "$INFORMATION", confirmation_message)
                if resp is not None and resp == 0:
                    _apply_binaries_update(pos_id)
                    return False

            elif pending_updates == "UPDATE_NOW":
                confirmation_message = "$HAS_PENDING_UPDATES_TO_APPLY_NOW"
                show_message_dialog(pos_id, "$INFORMATION", confirmation_message)
                _apply_binaries_update(pos_id)

    except Exception as ex:
        show_message_dialog(pos_id, "$ERROR", "$ERROR_CLOSING_DAY|{}".format(ex.message))


def _has_pending_pos_update(pos_id):
    msg = send_message("PosUpdater", TK_POS_HAS_BINARY_UPDATE_TO_APPLY, timeout=60 * 1000000)
    if msg.token != TK_SYS_ACK:
        sys_log_error("Error getting update binaries status on pos id: {}. Error: {}".format(pos_id, msg.data))

    return msg.data


def _apply_binaries_update(pos_id):
    msg = send_message("PosUpdater", TK_POS_UPDATER_PERFORM_BINARIES_UPDATE, FM_PARAM, "updateNow",
                       timeout=60 * 1000000)
    if msg.token == TK_SYS_ACK:
        return msg.data
    else:
        sys_log_error("Error applying update binaries status on pos id: {}. Error: {}".format(pos_id, msg.data))


def _fill_pos_list_with_can_close_day_pos(error_dict, period, pos_list, opened_users):
    for pos_no in sorted(pos_list):
        model = get_model(pos_no)
        if not _verify_if_can_close_day(pos_no, model, period, error_dict, opened_users):
            pos_list.remove(pos_no)


def _parse_error_messages(model, error_dict):
    error_messages = []
    if len(error_dict) > 0:
        for key in error_dict:
            if key in MwSysErrors.list_values():
                msg = translate_message(model, MwSysErrors.get_name(key))
            else:
                msg = str(key)

            error_messages.append("POS {}".format(error_dict[key]) + (" - {}".format(msg)))
    return error_messages


def _perform_close_day(pos_id, succeed_pos, error_dict):
    try:
        sys_log_info("Closing business day on POS id: [{}]".format(pos_id))

        pos_model = get_model(pos_id)
        if pos_model.find("PosState").get("state") == "CLOSED":
            return

        msg = send_message("POS%d" % int(pos_id), TK_POS_BUSINESSEND, FM_PARAM, str(pos_id), timeout=60 * 1000000)
        if msg.token == TK_SYS_ACK:
            succeed_pos.append(pos_id)
            return
        else:
            sys_log_error("Error closing business day on pos id: {}. Error: {}".format(pos_id, msg.data))
            _fill_error_dict(error_dict, pos_id, msg.data)

    except Exception as ex:
        sys_log_exception("Error closing business day on pos id: {}".format(pos_id))
        _fill_error_dict(error_dict, pos_id, ex.message)


def _verify_if_can_close_day(pos_id, model, period, error_dict, opened_users):
    try:
        if model.find("PosState").get("state") == "CLOSED":
            error_msg = translate_message(model, "DAY_IS_ALREADY_CLOSED")
            _fill_error_dict(error_dict, pos_id, error_msg)
            return False

        operator = model.find("Operator")
        if operator is not None and operator.get('id') in [x.get('id') for x in opened_users]:
            error_msg = translate_message(model, "THERE_ARE_OPEN_OPERATORS")
            _fill_error_dict(error_dict, pos_id, error_msg)
            return False

        if period == '0':
            error_msg = translate_message(model, "NEED_TO_OPEN_DAY_FIRST")
            _fill_error_dict(error_dict, pos_id, error_msg)
            return False
    except Exception as ex:
        _fill_error_dict(error_dict, pos_id, ex.message)

    return True


def _fill_error_dict(error_dict, pos_id, error_message):
    error_message = error_message.split("\0")[0]

    if error_message not in error_dict:
        error_dict[error_message] = [pos_id]
    else:
        error_dict[error_message].append(pos_id)

    return error_dict



# encoding: utf-8
from mw import MwOperatorNameRetriever

from openedtablesreport import OpenedTablesReport
from openedtablesreport.repository import TableRepository


class ReportGenerator(object):
    @staticmethod
    def generate(mb_context, pos_id):
        operator_retriever = MwOperatorNameRetriever(mb_context)
        table_repository = TableRepository(mb_context, pos_id)
        opened_tables_report_generator = OpenedTablesReport(table_repository, pos_id, operator_retriever)
        return opened_tables_report_generator.generate_opened_tables_report(pos_id)




# coding: utf-8
"""
Fixer for the execfile() function on Py2, which was removed in Py3.

The Lib/lib2to3/fixes/fix_execfile.py module has some problems: see
python-future issue #37. This fixer merely imports execfile() from
past.builtins and leaves the code alone.

Adds this import line::

    from past.builtins import execfile

for the function execfile() that was removed from Py3.
"""

from __future__ import unicode_literals
from lib2to3 import fixer_base

from libfuturize.fixer_util import touch_import_top


expression = "name='execfile'"


class FixExecfile(fixer_base.BaseFix):
    BM_compatible = True
    run_order = 9

    PATTERN = """
              power<
                 ({0}) trailer< '(' args=[any] ')' >
              rest=any* >
              """.format(expression)

    def transform(self, node, results):
        name = results["name"]
        touch_import_top(u'past.builtins', name.value, node)




# -*- coding: utf-8 -*-
import collections
import time
import xml.sax
import xml.sax.handler
from cStringIO import StringIO
from xml.etree import ElementTree as eTree
import cfgtools
import os

# START - DO NOT CHANGE THIS IMPORT ORDINATION
import constants
import sysactions
from helper import MwOrderStatus, remove_accents
from production.processor import SystemOrderParser, OrderXml
from reports import mbcontext
import printer_wrapper
from constants import *
from systools import sys_log_exception, sys_log_warning, sys_parsefilepath

from picklistreport.repository import ProductRepository
# END - DO NOT CHANGE THIS IMPORT ORDINATION

sysactions.mbcontext = mbcontext

COLS = 42
SMALL_COLS = 42
SIMPLE_SEPARATOR = ("-" * SMALL_COLS)
SEPARATOR = ("=" * COLS)
SMALL_SEPARATOR = ("=" * SMALL_COLS)
DATE_TIME_FMT = "%d/%m/%Y %H:%M:%S"
DATE_FMT = "%d/%m/%Y"
TIME_FMT = "%H:%M:%S"
PICKUP_TYPE = None
CUSTOMER_TEXT = "\nCliente.......: {}"

config = cfgtools.read(os.environ["LOADERCFG"])
delivery_address = sys_parsefilepath(config.find_value("Picklist.DeliveryAddress") or "true").lower() == "true"
picklist_font_size_items = None


def production_pick_list(order_xml, *args):
    try:
        picklist_name = args[0] if len(args) > 0 else ""

        production_order, is_reprint = _parse_order_xml_to_production_order(order_xml)
        if _order_is_voided(production_order):
            sys_log_warning("ignoring voided order")
            return Exception()

        pod_type = production_order.get("pod_type")
        if pod_type == "KK":
            return Exception()

        report = StringIO()

        report.write("%(TXT_ALIGN_LT)s" % _join(globals(), locals()))
        report.write("%(TXT_BOLD_ON)s" % _join(globals(), locals()))

        report_header = _get_report_header(production_order, is_reprint, picklist_name)
        report.write(report_header)

        report.write("%(TXT_BOLD_OFF)s" % _join(globals(), locals()))

        pre_sale_lines_report = report.tell()

        formatted_sale_lines = _get_formatted_sale_lines_by_seats(production_order)

        report.write(formatted_sale_lines)

        if pre_sale_lines_report == report.tell():
            sys_log_warning("Order has no sale lines to print")
            return Exception()

        report = _add_additional_comments(report, production_order)
        report = _add_sale_type(report, production_order)

        qr_code = _get_qr_code(production_order)
        if qr_code:
            report.write(qr_code)

        return report.getvalue()
    except Exception as _:
        sys_log_exception("Error building production PickList")


def _order_is_voided(production_order):
    return MwOrderStatus.VOIDED.name == production_order.get("state", "")


def _get_qr_code(production_order):
    global PICKUP_TYPE
    try:
        if PICKUP_TYPE is None:
            PICKUP_TYPE = sysactions.get_storewide_config("Store.pickUpDisplayType", defval="Manager")
    except Exception as _:
        PICKUP_TYPE = 'Manager'

    # Prints QRCode only when it will be used for PickupDisplay on PickList
    if "PickList" in PICKUP_TYPE:
        return _create_qrcode('QR' + (production_order.get("order_id")).zfill(8) + 'BK')
    else:
        return ""


def _parse_order_xml_to_production_order(order_xml):
    order_xml = eTree.XML(order_xml)
    if order_xml.tag == "Orders":
        order = order_xml.find('Order')
        product_repository = ProductRepository(mbcontext)
        part_code_jit_lines = product_repository.get_all_part_code_jit_lines()
        handler = SystemOrderParser(part_code_jit_lines)
        xml.sax.parseString(eTree.tostring(order), handler)
        order_xml = handler.order
        convert_xml = OrderXml()
        order_xml = eTree.XML(convert_xml.to_xml(order_xml))
    production_order = order_xml if order_xml.tag == "ProductionOrder" else order_xml.find("ProductionOrder")

    is_reprint = False
    production_items = production_order.findall(".//Items") or []
    for item in production_items:
        is_reprint = len(item.findall(".//TagHistory/TagEvent[@tag='printed']")) > 0

    return production_order, is_reprint


def _get_report_header(production_order, is_reprint, picklist_name):
    header_pod_type = production_order.get("pod_type")
    header_sale_type = production_order.get("sale_type")
    totem_print = header_pod_type == "TT" or header_sale_type == "TOTEM"
    drive_print = header_pod_type == "DT" or header_sale_type == "DRIVE_THRU"
    delivery_print = header_pod_type == "DL" or header_sale_type == "DELIVERY"
    sale_type_text, sale_type = _get_sale_type(delivery_print, drive_print, production_order, totem_print)

    title = _center(picklist_name + sale_type_text, SMALL_COLS)

    produced_datetime = ""
    timestamp = _get_first_printed_time(production_order)
    if timestamp is not None:
        timestamp = time.strptime(timestamp[:19], "%Y-%m-%dT%H:%M:%S")
        produced_datetime = time.strftime(DATE_TIME_FMT, timestamp)

    posid = int(production_order.get("originator")[-2:])
    order_id = production_order.get("order_id")
    if order_id is not None:
        order_id = order_id.split("-")[0]
    user_id = production_order.get("session_id").split("user=")[1].split(",")[0]
    user_name = eTree.XML(sysactions.get_user_information(user_id)).find(".//user").get("LongName")[:15]
    user_info = "{} / {}".format(user_id, user_name)

    order_id = int(order_id) % 10000
    table_name = None
    if production_order.find(".//Property[@key='TAB_ID']") is not None:
        order_descr = "Comanda"
        table_name = production_order.find(".//Property[@key='TAB_ID']").get('value')
        table_name = table_name[:5] + "..." if len(table_name) > 5 else table_name
        order_descr_id = "{}".format(table_name)
    elif production_order.find(".//Property[@key='TABLE_NAME']") is not None:
        order_descr = "Mesa"
        table_name = production_order.find(".//Property[@key='TABLE_NAME']").get('value')
        table_name = table_name[:6] + "..." if len(table_name) > 6 else table_name
        order_descr_id = "{}".format(table_name)
    else:
        order_descr = "Pedido"
        order_descr_id = "{}".format(order_id)

    created_at = time.strptime(production_order.get("created_at"), "%Y-%m-%dT%H:%M:%S.%f")
    created_at = time.strftime(DATE_TIME_FMT, created_at)

    preheader = ""
    if is_reprint:
        preheader = "*** REIMPRESSO ***\n"
    preheader = _center(preheader, SMALL_COLS)

    header = constants.TXT_SIZE + chr(constants.TXT_WIDTH[2] + constants.TXT_HEIGHT[2])
    header += _center("%(order_descr)s: %(order_descr_id)s" % _join(globals(), locals()), COLS/2)
    header += "%(TXT_NORMAL)s%(TXT_BOLD_OFF)s%(TXT_FONT_A)s\n" % _join(globals(), locals())
    header += "%(preheader)s%(SMALL_SEPARATOR)s\n%(title)s" % _join(globals(), locals())
    header += "\nCodigo........: %(order_id)s (POS #%(posid)02d)" % _join(globals(), locals())
    header += _print_delivery_partner(delivery_print, production_order)

    customer_name = _get_xml_node_value(production_order, "CUSTOMER_NAME")
    if customer_name:
        header += CUSTOMER_TEXT.format(customer_name)

    current_datetime = time.strftime(DATE_TIME_FMT, time.localtime())
    if produced_datetime:
        header += "\nHora Pedido...: {}".format(created_at)
        header += "\nHora Produzido: {}".format(produced_datetime)
    else:
        header += "\nImpresso em...: {}".format(current_datetime)

    if is_reprint:
        header += "\nReimpresso....: %s" % current_datetime

    header += "\nOperador......: {}".format(user_info)
    header += "\n{}".format(SMALL_SEPARATOR)

    header += _print_delivery_info(delivery_print, production_order, sale_type)
    return header


def _print_delivery_partner(delivery_print, production_order):
    text = ""
    partner_node = _get_xml_node_value(production_order, "PARTNER")
    if delivery_print and partner_node:
        partner = partner_node.get("value", "") if hasattr(partner_node, 'get') else partner_node

        partner_id = ""
        partner_id_node = _get_xml_node_value(production_order, "SHORT_REFERENCE")
        if partner_id_node is not None:
            partner_id = partner_id_node.get("value") if hasattr(partner_id_node, 'get') else partner_id_node

        if partner != "" and partner_id != "":
            text += """\nParceiro......: {0}\nID Parceiro...: {1}""".format(partner.upper(), partner_id)

    return text


def _print_delivery_info(delivery_print, production_order, sale_type):
    text = ""
    valid_sale_type = sale_type is not None and sale_type.upper() not in ["TAKE_OUT", "EAT_IN"]
    if not delivery_print or not delivery_address or not valid_sale_type:
        return text

    try:
        street_name_element = _get_xml_node_value(production_order, "STREET_NAME")
        street_number_element = _get_xml_node_value(production_order, "STREET_NUMBER")
        if street_name_element:
            text += """\nEndereco....: {0}, {1}\n""".format(street_name_element, street_number_element)

        text += get_node_value(production_order, "COMPLEMENT", "Complemento.")
        text += get_node_value(production_order, "REFERENCE", "Referencia..")
        text += get_node_value(production_order, "NEIGHBORHOOD", "Bairro......")
        text += get_node_value(production_order, "POSTAL_CODE", "CEP.........")
        text += get_node_value(production_order, "CITY", "Cidade......")
        text += get_node_value(production_order, "STATE", "Estado......")
        if text != "":
            text += SMALL_SEPARATOR

    except Exception:
        sys_log_exception("Error in get address on Delivery picklist")

    return text


def get_node_value(production_order, node_value, text):
    complement_element = _get_xml_node_value(production_order, node_value)
    if complement_element:
        return """{}: {}\n""".format(text, complement_element)
    return ""


def _get_sale_type(delivery_print, drive_print, production_order, totem_print):
    sale_type_node = ""

    if totem_print:
        return "(TOTEM)", sale_type_node
    if drive_print:
        return "(DRIVE-THRU)", sale_type_node
    if delivery_print:
        sale_type, sale_type_node = _get_sale_type_by_pickup_type(production_order)
        return sale_type, sale_type_node

    sale_type, sale_type_node = _get_by_sale_type(production_order)

    return sale_type, sale_type_node


def _get_sale_type_by_pickup_type(production_order):
    sale_type = " (DELIVERY)"

    sale_type_node = _get_xml_node_value(production_order, "PICKUP_TYPE")
    if sale_type_node is not None:
        sale_type_node = sale_type_node.get("value", "") if hasattr(sale_type_node, 'get') else sale_type_node
        if sale_type_node.upper() == "TAKE_OUT":
            sale_type = " (DELIVERY - RETIRADA VIAGEM)"
        elif sale_type_node.upper() == "EAT_IN":
            sale_type = " (DELIVERY - RETIRADA LOJA)"

    return sale_type, sale_type_node


def _get_by_sale_type(production_order):
    sale_type_node = _get_xml_node_value(production_order, "SALE_TYPE")
    sale_type_node = sale_type_node.get("value", "") if hasattr(sale_type_node, 'get') else sale_type_node
    if sale_type_node is not None:
        if sale_type_node == "TAKE_OUT":
            sale_type = " (VIAGEM)"
        else:
            sale_type = " (LOJA)"
    else:
        sale_type = " (LOJA)"
        sale_type_node = ""

    return sale_type, sale_type_node


def _print_customer_name(production_order):
    # type: (eTree.XML) -> str

    customer_name = _get_xml_node_value(production_order, "CUSTOMER_NAME")
    partner_name = _get_xml_node_value(production_order, "PARTNER")
    short_reference = _get_xml_node_value(production_order, "SHORT_REFERENCE")

    if partner_name and short_reference:
        return CUSTOMER_TEXT.format(partner_name.upper() + " - " + short_reference)
    elif customer_name:
        return CUSTOMER_TEXT.format(customer_name.upper())

    return ""


def _get_grouped_sale_lines_by_seat(production_order):
    pos_id = int(production_order.get("originator")[-2:])
    model = sysactions.get_model(pos_id)

    grouped_sale_lines = {}
    pre_option_level = None
    preceding_removed_combo = False

    _fix_sub_items_seat(production_order)
    for item in production_order.findall(".//Item"):
        sale_line, pre_option_level, preceding_removed_combo = _get_formatted_sale_line(
                model,
                item,
                pre_option_level,
                preceding_removed_combo,
        )
        if sale_line:
            item_seat = item.get("seat") or 0
            if item_seat in grouped_sale_lines:
                grouped_sale_lines[item_seat].append(sale_line)
            else:
                grouped_sale_lines[item_seat] = [sale_line]

    return grouped_sale_lines.items() if len(grouped_sale_lines) > 0 else grouped_sale_lines


def _fix_sub_items_seat(production_order):
    items = production_order.findall("Items/Item")
    for item in items:
        item_sons = item.findall(".//Item")
        for son in item_sons:
            son.set("seat", item.get("seat"))


def _get_formatted_sale_lines_by_seats(production_order):

    formatted_sale_lines = "\n"
    formatted_sale_lines += _get_picklist_font_size_items()
    grouped_sale_lines = _get_grouped_sale_lines_by_seat(production_order)
    for seat, sale_lines_by_seat in collections.OrderedDict(sorted(grouped_sale_lines)).items():
        if production_order.find("./Properties/Property[@key='TABLE_ID']") is not None:
            formatted_sale_lines += _get_title_sale_lines_seats(seat, production_order)

        for sale_line in sale_lines_by_seat:
            if sale_line is not None:
                formatted_sale_lines += sale_line

        formatted_sale_lines += "\n"

    formatted_sale_lines += _rollback_printer_default_font()
    return formatted_sale_lines


def _get_title_sale_lines_seats(seat, production_order):
    # type: (str, eTree) -> str

    seat_title = " Assento {}"
    table_title = " Mesa"

    if int(seat) != 0:
        title = seat_title.format(seat)
    else:
        seats_number = production_order.find("./Properties/Property[@key='TABLE_SEATS']")
        seat_detail = " ({} assentos)".format(seats_number.get("value")) if seats_number is not None else ""
        title = table_title + seat_detail

    return "{}\n{}\n".format(title, SIMPLE_SEPARATOR)


def _get_formatted_sale_line(model, sale_line, pre_option_level=None, preceding_removed_combo=False, cols=COLS):
    item_type = None if "item_type" not in sale_line.attrib else sale_line.get("item_type")

    level = int(sale_line.get("level"))

    if item_type == "COMBO":
        preceding_removed_combo = False

    if level == 0:
        pre_option_level = None

    if item_type in [None, "OPTION", "COUPON"]:
        pre_option_level = level if item_type == "OPTION" else None
        return None, pre_option_level, preceding_removed_combo

    qty = float(sale_line.get("multiplied_qty") or sale_line.get("qty") or 0)
    default_qty = float(sale_line.get("default_qty") or 0)

    if (qty == 0 and default_qty == 0) or preceding_removed_combo:
        if item_type == "COMBO":
            preceding_removed_combo = True
        return None, None, preceding_removed_combo

    if qty < default_qty and qty == 0:
        qty_description = sysactions.translate_message(model, "WITHOUT")
    else:
        measure_unit = (sale_line.attrib.get("measure_unit", "UN")).upper()
        decimal_places = 3 if measure_unit == "KG" else 0
        quantity_format = "{" + ":.{}f".format(decimal_places) + "}"
        qty_description = quantity_format.format(qty)

    indentation = " " * ((pre_option_level or level) + 1)
    product_name = sale_line.get("description")

    manual_course = sale_line.get("manual_course")
    if manual_course and level == 0:
        manual_course = "[" + manual_course + "]"
    description_length = cols - len(manual_course)

    item_comment = sale_line.find("./Comment")
    if item_comment is not None:
        comment = item_comment.get("comment")

        comment_description = ""
        jumped_comment = ""
        if comment:
            if comment.startswith("["):
                comment_description = comment.replace("[", "").replace("]", "").upper()
                comment_description = sysactions.translate_message(model, comment_description)
            else:
                description = "[" + comment + "]"
                line_rjust = int(len(description) + len(indentation) + round(qty, 1))
                jumped_comment = remove_accents(description.rjust(line_rjust))

        product_description = str(comment_description or qty_description) + " " + product_name
        line_rjust = len(product_description) + len(indentation)

        product_description = product_description.rjust(line_rjust)
        if manual_course:
            product_description = product_description.ljust(description_length)[:description_length]

        description = remove_accents(product_description + manual_course)

        formatted_sale_line = description + "\n"
        if jumped_comment:
            formatted_sale_line += jumped_comment + "\n"
    else:
        product_description = indentation + qty_description + " " + product_name
        if manual_course:
            product_description = product_description.ljust(description_length)[:description_length]
        formatted_sale_line = remove_accents(product_description + manual_course) + "\n"

    if level <= pre_option_level:
        pre_option_level = None

    return formatted_sale_line, pre_option_level, preceding_removed_combo


def _get_xml_node_value(order_temp, prop_value, prop_name="key", value_name="value"):
    node = order_temp.findall(".//*[@{0}='{1}']".format(prop_name, prop_value))
    if node is not None and len(node) > 0:
        return remove_accents(node[0].get(value_name, ""))
    else:
        return ""


def _add_sale_type(report, order_t):
    sale_pod_type = order_t.get("pod_type")

    sale_type = ""
    if sale_pod_type == "DT":
        sale_type = "DRIVE-THRU"
    else:
        if order_t.get("sale_type") != 'EAT_IN':
            sale_type = (order_t.get("sale_type")).replace('_', ' ').upper()
            if sale_type == "APP":
                pickup_type_element = _get_xml_node_value(order_t, "PICKUP_TYPE").upper()

                if pickup_type_element is not None:
                    pickup_type_element = pickup_type_element.get("value", "") \
                        if hasattr(pickup_type_element, 'get') else pickup_type_element
                    if pickup_type_element.upper() == "TAKE_OUT":
                        sale_type = "VIAGEM"

                    if pickup_type_element.upper() == "EAT_IN":
                        sale_type = ""

            if "TAKE" in sale_type:
                sale_type = "VIAGEM"

    if sale_type != "":
        sale_type_text = constants.TXT_4SQUARE + _center_text(sale_type, '*', COLS / 2) + constants.TXT_NORMAL
        report.write("\n%s" % sale_type_text)

    return report


def _create_qrcode(qr_data):
    qr = qrcode.QRCode(
        version=2,
        error_correction=qrcode.constants.ERROR_CORRECT_H,
        box_size=5,
        border=4,
    )

    qr.add_data(qr_data)
    qr.make(fit=True)

    img = qr.make_image()
    img = img.convert("RGB")
    img = printer_wrapper.image(img)

    return img.strip()


def _add_additional_comments(report, production_order):
    tab_add_info = production_order.findall(".//Property[@key='TAB_ADD_INFO']")
    if tab_add_info and len(tab_add_info) > 0:
        tab_add_info_text = tab_add_info[0].attrib["value"]
        report.write("\n{}".format(SMALL_SEPARATOR))

        if len(tab_add_info_text) > COLS:
            report.write("\n\n%s" % remove_accents(tab_add_info_text[:COLS]))
            report.write("\n%s" % remove_accents(tab_add_info_text[COLS:2*COLS]))
        else:
            report.write("\n\n%s" % remove_accents(tab_add_info_text))

        report.write("\n\n{}".format(SMALL_SEPARATOR))

    return report


def _join(dic1, dic2):
    d = dict(dic1)
    d.update(dic2)
    return d


def _cut(s):
    if len(s) > COLS:
        # If there is a line-feed at the end, keep it
        return s[:COLS] if (s[-1] != "\n") else (s[:(COLS - 1)] + "\n")
    return s


def _center(s, cols=COLS):
    if not s:
        return s
    s = _cut(s)
    miss = cols - len(s)
    if miss == 0:
        return s
    left = miss / 2
    return (" " * left) + s


def _center_text(text, sep=" ", line_size=COLS):
    text = " %s " % text
    inicio = sep * int((line_size - len(text)) / 2)
    meio = inicio + text
    fim = sep * (line_size - len(meio))

    return meio + fim


def _get_first_printed_time(production_order):
    production_items = production_order.findall(".//Items") or []
    for item in production_items:
        serverd_tags = item.findall(".//TagHistory/TagEvent[@tag='served']") or []
        if len(serverd_tags) > 0:
            return serverd_tags[0].attrib["timestamp"]

    return None


def _get_picklist_font_size_items():
    # type: () -> str

    global picklist_font_size_items

    if picklist_font_size_items is None:
        small = constants.TXT_FONT_A
        large = constants.TXT_SIZE + chr(constants.TXT_WIDTH[2] + constants.TXT_HEIGHT[2])
        medium = constants.TXT_FONT_SELECT_C + constants.TXT_FONT_C
        try:
            sizes = {"small": small, "medium": medium, "large": large}
            font_size_items = str((config.find_value("Picklist.FontSize") or "small")).strip()
            picklist_font_size_items = sizes[font_size_items]
        except Exception as _:
            picklist_font_size_items = small

    return picklist_font_size_items


def _rollback_printer_default_font():
    # type: () -> str

    return "{}{}{}{}".format(constants.TXT_NORMAL, constants.TXT_BOLD_OFF, constants.TXT_FONT_A, constants.TXT_ALIGN_LT)



# encoding: utf-8
from datetime import datetime
from typing import List, Any, Dict

from basereport import SimpleReport, ReportColumnDefinition, TableReport, AlignTypes
from benefitsreport.model import OrderBenefit
from report import Presenter, Part
from report.command import NewLineCommand, AlignCommand, RepeatCommand, FontCommand
from reportsdata.benefitsreport.builders import BenefitConsolidatedBuilder
from utils.report_formatters import format_line, remove_accents, format_price

from old_helper import convert_from_utf_to_localtime
from systools import sys_log_exception


class BenefitsPresenter(Presenter):
    def __init__(self, columns):
        # type: (int) -> None

        self.columns = columns
        self.header_column_definition = [
            ReportColumnDefinition(width=16, after_fill_text=': ', fill_with_char='.'),
            ReportColumnDefinition(width=columns - 16)
        ]

        self.period_column_definition = [
            ReportColumnDefinition(width=10, after_fill_text=' - ', fill_with_char=''),
            ReportColumnDefinition(width=columns - 10)
        ]

        self.header_table_column_definitions = [
            ReportColumnDefinition(width=7, after_text=' ', fill_with_char='', align=AlignTypes.LEFT),
            ReportColumnDefinition(width=12, after_text=' ', fill_with_char=''),
            ReportColumnDefinition(width=4, after_text=' ', fill_with_char='', align=AlignTypes.CENTER),
            ReportColumnDefinition(width=8),
            ReportColumnDefinition(width=7),
        ]

        self.sub_header_table_column_definitions = [
            ReportColumnDefinition(width=7),
            ReportColumnDefinition(width=12, after_text='', fill_with_char=''),
            ReportColumnDefinition(width=4, after_text='', fill_with_char='', align=AlignTypes.CENTER),
            ReportColumnDefinition(width=7, align=AlignTypes.RIGHT),
            ReportColumnDefinition(width=8, align=AlignTypes.RIGHT),
        ]

        self.items_table_column_definitions = [
            ReportColumnDefinition(width=7),
            ReportColumnDefinition(width=12),
            ReportColumnDefinition(width=4, after_text='', fill_with_char='', align=AlignTypes.CENTER),
            ReportColumnDefinition(width=7, align=AlignTypes.RIGHT),
            ReportColumnDefinition(width=8, align=AlignTypes.RIGHT),
        ]

        self.header_table_column_consolidated_definitions = [
            ReportColumnDefinition(width=16, align=AlignTypes.CENTER, fill_with_char=''),
            ReportColumnDefinition(width=4, align=AlignTypes.CENTER, fill_with_char=''),
            ReportColumnDefinition(width=18, align=AlignTypes.CENTER, fill_with_char=''),
            ReportColumnDefinition(width=18, align=AlignTypes.CENTER, fill_with_char=''),
        ]

    def present(self, dto):
        try:
            parts = [Part(commands=[FontCommand(FontCommand.Font.B)])]
            parts.extend(self._build_header(dto.header))
            parts.extend(self._build_items(dto.body))
            parts.extend([Part(None, [NewLineCommand(), NewLineCommand()])])
        except Exception as _:
            sys_log_exception("Error building Benefits Report")
            raise

        return SimpleReport(parts, self.columns)

    def _build_header(self, header):
        parts = [
            Part(None, [NewLineCommand()]),
            self._get_title(header),
            Part(None, [NewLineCommand(), NewLineCommand()])
        ]
        parts.extend(self._get_promotion_name(header))
        parts.extend(self._get_store_name(header))
        parts.extend(self._get_date_time())
        parts.extend(self._get_period(header))
        parts.extend(self._get_operator_name(header))
        parts.extend([Part(None, [RepeatCommand(self.columns, "="), NewLineCommand()])])

        return parts

    def _build_items(self, body):
        type_report = body.type_report
        items = body.items
        if type_report == "detailed":
            return self._build_items_detailed(items)

        return BenefitConsolidatedBuilder(body, self.columns).generate()

    def _build_items_detailed(self, lines):
        # type: (Dict) -> List

        data = []
        total_qty = 0
        total_default_price = 0.0
        total_discount_value = 0.0
        for item in lines:
            items = []
            qty = 0
            default_price = 0.0
            discount_value = 0.0
            orders = []
            total_products = []
            for product in lines[item]:
                data_table = self._build_product(product)
                self._append_list_header(data_table, orders, total_products)
                qty += float(data_table[2])
                default_price += float(data_table[3].replace(",", ""))
                discount_value += float(data_table[4].replace(",", ""))
                items.extend([TableReport([data_table], self.items_table_column_definitions)])

            orders = str(len(orders)).zfill(5)
            total_products = str(len(total_products)).zfill(3)
            formatted_default_price = format_price(default_price)
            formatted_discount_value = format_price(discount_value)
            totals = [orders, total_products, str(qty).zfill(4), formatted_default_price, formatted_discount_value]

            total_qty += qty
            total_default_price += default_price
            total_discount_value += discount_value
            data.extend(self._build_header_and_items(items, item, totals))

        total_default_price = format_price(total_default_price)
        total_discount_value = format_price(total_discount_value)
        grand_total = [str(total_qty).zfill(4), total_default_price,  total_discount_value]
        data.extend(self._build_footer_table(grand_total))

        return data

    @staticmethod
    def _append_list_header(data_table, orders, total_products):

        if data_table[0] not in orders:
            orders.append(data_table[0])
        if data_table[1] not in total_products:
            total_products.append(data_table[1])

    def _build_header_and_items(self, items, item, totals):
        # type: (List, Any, List) -> List

        data = []
        period = datetime.strptime(item, '%Y%m%d').strftime('%d/%m/%Y')
        data.extend([Part(None, [NewLineCommand()])])
        data.extend(format_line("$BENEFIT_REPORT_PERIOD", period, self.columns, self.period_column_definition))
        header = self._build_header_table(totals)
        data.extend([Part(None, [NewLineCommand()])])
        data.extend(header)
        data.extend(items)
        data.extend([Part(None, [NewLineCommand()])])
        data.extend([Part(None, [RepeatCommand(self.columns, "-"), NewLineCommand(), NewLineCommand()])])

        return data

    def _build_header_table(self, totals):
        # type: (List) -> List

        header = [
            TableReport([["$BENEFIT_REPORT_HEADER_ORDER",
                          "$BENEFIT_REPORT_HEADER_ITEMS",
                          "$QTY",
                          "$BENEFIT_REPORT_HEADER_DEFAULT_VALUE",
                          "$BENEFIT_REPORT_HEADER_DISCOUNT_VALUE"]],
                        self.header_table_column_definitions),
            ]
        header.extend([TableReport([totals], self.sub_header_table_column_definitions)])
        header.extend([Part(None, [RepeatCommand(self.columns, "="), NewLineCommand()])])
        return header

    def _build_footer_table(self, totals):
        # type: (List) -> List

        return [TableReport([["$TOTAL", " ", totals[0], totals[1], totals[2]]], self.items_table_column_definitions)]

    @staticmethod
    def _build_product(product):
        # type: (OrderBenefit) -> List

        order_id = product.order_id.zfill(5)
        product_name = product.product_name
        default_price = format_price(product.ordered_qty * product.default_unit_price)
        discount_value = format_price(product.discount_value)

        return [order_id, product_name, str(product.ordered_qty).zfill(4), default_price, discount_value]

    @staticmethod
    def _get_title(header):
        return Part(header.title, [AlignCommand(AlignCommand.Alignment.center)])

    def _get_promotion_name(self, header):
        promotion_name = header.promotion_name
        title = "$BENEFIT_REPORT_PROMOTION_NAME"
        return format_line(title, promotion_name, self.columns, self.header_column_definition)

    def _get_store_name(self, header):
        return format_line("$STORE", header.store_name, self.columns, self.header_column_definition)

    def _get_date_time(self):
        current_datetime = convert_from_utf_to_localtime(datetime.now()).strftime('%d/%m/%y %H:%M:%S')
        return format_line("$BENEFIT_REPORT_DATE_TIME", current_datetime, self.columns, self.header_column_definition)

    def _get_period(self, header):
        period = "{0} - {1}".format(header.start_date, header.end_date)
        return format_line("$BENEFIT_REPORT_PERIOD", period, self.columns, self.header_column_definition)

    def _get_operator_name(self, header):
        operator = remove_accents(header.operator).upper()
        return format_line("$OPERATOR", operator, self.columns, self.header_column_definition)



from application.interactor import BurnBenefitInteractor, UnlockBenefitInteractor
from application.model import BenefitControllerDto, OperationDescription
from application.model.configuration import Configurations
from application.repository.pos import DBRepository
from typing import List, Optional


class BenefitRetryInteractor(object):
    
    def __init__(self, configs, db_repository, burn_benefit_interactor, unlock_benefit_interactor):
        # type: (Configurations, DBRepository, BurnBenefitInteractor, UnlockBenefitInteractor) -> None

        self.configs = configs
        self.logger = self.configs.logger
        self.db_repository = db_repository
        self.burn_interactor = burn_benefit_interactor
        self.unlock_interactor = unlock_benefit_interactor

    def get_pending_operations_benefits(self):
        # type: () -> Optional[List[BenefitControllerDto]]
        
        return self.db_repository.find_pending_operations()

    def perform_operations_on_pending_benefits(self, pending_operations):
        # type: (List[BenefitControllerDto]) -> None
        
        for benefit in pending_operations:
            default_params = [benefit.benefit_id,
                              benefit.originator_pos_id,
                              benefit.benefit_applier,
                              benefit.retry_quantity]
            
            if benefit.operation_description == OperationDescription.BURNED:
                self.burn_interactor.burn_benefit(*default_params)

            elif benefit.operation_description == OperationDescription.UNLOCKED:
                self.unlock_interactor.unlock_benefit(*default_params)



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import absolute_import, division, print_function

import abc

import six


@six.add_metaclass(abc.ABCMeta)
class KeyDerivationFunction(object):
    @abc.abstractmethod
    def derive(self, key_material):
        """
        Deterministically generates and returns a new key based on the existing
        key material.
        """

    @abc.abstractmethod
    def verify(self, key_material, expected_key):
        """
        Checks whether the key generated by the key material matches the
        expected derived key. Raises an exception if they do not match.
        """



# -*- coding: utf-8 -*-
"""
    flask.blueprints
    ~~~~~~~~~~~~~~~~

    Blueprints are the recommended way to implement larger or more
    pluggable applications in Flask 0.7 and later.

    :copyright: 2010 Pallets
    :license: BSD-3-Clause
"""
from functools import update_wrapper

from .helpers import _endpoint_from_view_func
from .helpers import _PackageBoundObject

# a singleton sentinel value for parameter defaults
_sentinel = object()


class BlueprintSetupState(object):
    """Temporary holder object for registering a blueprint with the
    application.  An instance of this class is created by the
    :meth:`~flask.Blueprint.make_setup_state` method and later passed
    to all register callback functions.
    """

    def __init__(self, blueprint, app, options, first_registration):
        #: a reference to the current application
        self.app = app

        #: a reference to the blueprint that created this setup state.
        self.blueprint = blueprint

        #: a dictionary with all options that were passed to the
        #: :meth:`~flask.Flask.register_blueprint` method.
        self.options = options

        #: as blueprints can be registered multiple times with the
        #: application and not everything wants to be registered
        #: multiple times on it, this attribute can be used to figure
        #: out if the blueprint was registered in the past already.
        self.first_registration = first_registration

        subdomain = self.options.get("subdomain")
        if subdomain is None:
            subdomain = self.blueprint.subdomain

        #: The subdomain that the blueprint should be active for, ``None``
        #: otherwise.
        self.subdomain = subdomain

        url_prefix = self.options.get("url_prefix")
        if url_prefix is None:
            url_prefix = self.blueprint.url_prefix
        #: The prefix that should be used for all URLs defined on the
        #: blueprint.
        self.url_prefix = url_prefix

        #: A dictionary with URL defaults that is added to each and every
        #: URL that was defined with the blueprint.
        self.url_defaults = dict(self.blueprint.url_values_defaults)
        self.url_defaults.update(self.options.get("url_defaults", ()))

    def add_url_rule(self, rule, endpoint=None, view_func=None, **options):
        """A helper method to register a rule (and optionally a view function)
        to the application.  The endpoint is automatically prefixed with the
        blueprint's name.
        """
        if self.url_prefix is not None:
            if rule:
                rule = "/".join((self.url_prefix.rstrip("/"), rule.lstrip("/")))
            else:
                rule = self.url_prefix
        options.setdefault("subdomain", self.subdomain)
        if endpoint is None:
            endpoint = _endpoint_from_view_func(view_func)
        defaults = self.url_defaults
        if "defaults" in options:
            defaults = dict(defaults, **options.pop("defaults"))
        self.app.add_url_rule(
            rule,
            "%s.%s" % (self.blueprint.name, endpoint),
            view_func,
            defaults=defaults,
            **options
        )


class Blueprint(_PackageBoundObject):
    """Represents a blueprint, a collection of routes and other
    app-related functions that can be registered on a real application
    later.

    A blueprint is an object that allows defining application functions
    without requiring an application object ahead of time. It uses the
    same decorators as :class:`~flask.Flask`, but defers the need for an
    application by recording them for later registration.

    Decorating a function with a blueprint creates a deferred function
    that is called with :class:`~flask.blueprints.BlueprintSetupState`
    when the blueprint is registered on an application.

    See :ref:`blueprints` for more information.

    .. versionchanged:: 1.1.0
        Blueprints have a ``cli`` group to register nested CLI commands.
        The ``cli_group`` parameter controls the name of the group under
        the ``flask`` command.

    .. versionadded:: 0.7

    :param name: The name of the blueprint. Will be prepended to each
        endpoint name.
    :param import_name: The name of the blueprint package, usually
        ``__name__``. This helps locate the ``root_path`` for the
        blueprint.
    :param static_folder: A folder with static files that should be
        served by the blueprint's static route. The path is relative to
        the blueprint's root path. Blueprint static files are disabled
        by default.
    :param static_url_path: The url to serve static files from.
        Defaults to ``static_folder``. If the blueprint does not have
        a ``url_prefix``, the app's static route will take precedence,
        and the blueprint's static files won't be accessible.
    :param template_folder: A folder with templates that should be added
        to the app's template search path. The path is relative to the
        blueprint's root path. Blueprint templates are disabled by
        default. Blueprint templates have a lower precedence than those
        in the app's templates folder.
    :param url_prefix: A path to prepend to all of the blueprint's URLs,
        to make them distinct from the rest of the app's routes.
    :param subdomain: A subdomain that blueprint routes will match on by
        default.
    :param url_defaults: A dict of default values that blueprint routes
        will receive by default.
    :param root_path: By default, the blueprint will automatically this
        based on ``import_name``. In certain situations this automatic
        detection can fail, so the path can be specified manually
        instead.
    """

    warn_on_modifications = False
    _got_registered_once = False

    #: Blueprint local JSON decoder class to use.
    #: Set to ``None`` to use the app's :class:`~flask.app.Flask.json_encoder`.
    json_encoder = None
    #: Blueprint local JSON decoder class to use.
    #: Set to ``None`` to use the app's :class:`~flask.app.Flask.json_decoder`.
    json_decoder = None

    # TODO remove the next three attrs when Sphinx :inherited-members: works
    # https://github.com/sphinx-doc/sphinx/issues/741

    #: The name of the package or module that this app belongs to. Do not
    #: change this once it is set by the constructor.
    import_name = None

    #: Location of the template files to be added to the template lookup.
    #: ``None`` if templates should not be added.
    template_folder = None

    #: Absolute path to the package on the filesystem. Used to look up
    #: resources contained in the package.
    root_path = None

    def __init__(
        self,
        name,
        import_name,
        static_folder=None,
        static_url_path=None,
        template_folder=None,
        url_prefix=None,
        subdomain=None,
        url_defaults=None,
        root_path=None,
        cli_group=_sentinel,
    ):
        _PackageBoundObject.__init__(
            self, import_name, template_folder, root_path=root_path
        )
        self.name = name
        self.url_prefix = url_prefix
        self.subdomain = subdomain
        self.static_folder = static_folder
        self.static_url_path = static_url_path
        self.deferred_functions = []
        if url_defaults is None:
            url_defaults = {}
        self.url_values_defaults = url_defaults
        self.cli_group = cli_group

    def record(self, func):
        """Registers a function that is called when the blueprint is
        registered on the application.  This function is called with the
        state as argument as returned by the :meth:`make_setup_state`
        method.
        """
        if self._got_registered_once and self.warn_on_modifications:
            from warnings import warn

            warn(
                Warning(
                    "The blueprint was already registered once "
                    "but is getting modified now.  These changes "
                    "will not show up."
                )
            )
        self.deferred_functions.append(func)

    def record_once(self, func):
        """Works like :meth:`record` but wraps the function in another
        function that will ensure the function is only called once.  If the
        blueprint is registered a second time on the application, the
        function passed is not called.
        """

        def wrapper(state):
            if state.first_registration:
                func(state)

        return self.record(update_wrapper(wrapper, func))

    def make_setup_state(self, app, options, first_registration=False):
        """Creates an instance of :meth:`~flask.blueprints.BlueprintSetupState`
        object that is later passed to the register callback functions.
        Subclasses can override this to return a subclass of the setup state.
        """
        return BlueprintSetupState(self, app, options, first_registration)

    def register(self, app, options, first_registration=False):
        """Called by :meth:`Flask.register_blueprint` to register all views
        and callbacks registered on the blueprint with the application. Creates
        a :class:`.BlueprintSetupState` and calls each :meth:`record` callback
        with it.

        :param app: The application this blueprint is being registered with.
        :param options: Keyword arguments forwarded from
            :meth:`~Flask.register_blueprint`.
        :param first_registration: Whether this is the first time this
            blueprint has been registered on the application.
        """
        self._got_registered_once = True
        state = self.make_setup_state(app, options, first_registration)

        if self.has_static_folder:
            state.add_url_rule(
                self.static_url_path + "/<path:filename>",
                view_func=self.send_static_file,
                endpoint="static",
            )

        for deferred in self.deferred_functions:
            deferred(state)

        cli_resolved_group = options.get("cli_group", self.cli_group)

        if not self.cli.commands:
            return

        if cli_resolved_group is None:
            app.cli.commands.update(self.cli.commands)
        elif cli_resolved_group is _sentinel:
            self.cli.name = self.name
            app.cli.add_command(self.cli)
        else:
            self.cli.name = cli_resolved_group
            app.cli.add_command(self.cli)

    def route(self, rule, **options):
        """Like :meth:`Flask.route` but for a blueprint.  The endpoint for the
        :func:`url_for` function is prefixed with the name of the blueprint.
        """

        def decorator(f):
            endpoint = options.pop("endpoint", f.__name__)
            self.add_url_rule(rule, endpoint, f, **options)
            return f

        return decorator

    def add_url_rule(self, rule, endpoint=None, view_func=None, **options):
        """Like :meth:`Flask.add_url_rule` but for a blueprint.  The endpoint for
        the :func:`url_for` function is prefixed with the name of the blueprint.
        """
        if endpoint:
            assert "." not in endpoint, "Blueprint endpoints should not contain dots"
        if view_func and hasattr(view_func, "__name__"):
            assert (
                "." not in view_func.__name__
            ), "Blueprint view function name should not contain dots"
        self.record(lambda s: s.add_url_rule(rule, endpoint, view_func, **options))

    def endpoint(self, endpoint):
        """Like :meth:`Flask.endpoint` but for a blueprint.  This does not
        prefix the endpoint with the blueprint name, this has to be done
        explicitly by the user of this method.  If the endpoint is prefixed
        with a `.` it will be registered to the current blueprint, otherwise
        it's an application independent endpoint.
        """

        def decorator(f):
            def register_endpoint(state):
                state.app.view_functions[endpoint] = f

            self.record_once(register_endpoint)
            return f

        return decorator

    def app_template_filter(self, name=None):
        """Register a custom template filter, available application wide.  Like
        :meth:`Flask.template_filter` but for a blueprint.

        :param name: the optional name of the filter, otherwise the
                     function name will be used.
        """

        def decorator(f):
            self.add_app_template_filter(f, name=name)
            return f

        return decorator

    def add_app_template_filter(self, f, name=None):
        """Register a custom template filter, available application wide.  Like
        :meth:`Flask.add_template_filter` but for a blueprint.  Works exactly
        like the :meth:`app_template_filter` decorator.

        :param name: the optional name of the filter, otherwise the
                     function name will be used.
        """

        def register_template(state):
            state.app.jinja_env.filters[name or f.__name__] = f

        self.record_once(register_template)

    def app_template_test(self, name=None):
        """Register a custom template test, available application wide.  Like
        :meth:`Flask.template_test` but for a blueprint.

        .. versionadded:: 0.10

        :param name: the optional name of the test, otherwise the
                     function name will be used.
        """

        def decorator(f):
            self.add_app_template_test(f, name=name)
            return f

        return decorator

    def add_app_template_test(self, f, name=None):
        """Register a custom template test, available application wide.  Like
        :meth:`Flask.add_template_test` but for a blueprint.  Works exactly
        like the :meth:`app_template_test` decorator.

        .. versionadded:: 0.10

        :param name: the optional name of the test, otherwise the
                     function name will be used.
        """

        def register_template(state):
            state.app.jinja_env.tests[name or f.__name__] = f

        self.record_once(register_template)

    def app_template_global(self, name=None):
        """Register a custom template global, available application wide.  Like
        :meth:`Flask.template_global` but for a blueprint.

        .. versionadded:: 0.10

        :param name: the optional name of the global, otherwise the
                     function name will be used.
        """

        def decorator(f):
            self.add_app_template_global(f, name=name)
            return f

        return decorator

    def add_app_template_global(self, f, name=None):
        """Register a custom template global, available application wide.  Like
        :meth:`Flask.add_template_global` but for a blueprint.  Works exactly
        like the :meth:`app_template_global` decorator.

        .. versionadded:: 0.10

        :param name: the optional name of the global, otherwise the
                     function name will be used.
        """

        def register_template(state):
            state.app.jinja_env.globals[name or f.__name__] = f

        self.record_once(register_template)

    def before_request(self, f):
        """Like :meth:`Flask.before_request` but for a blueprint.  This function
        is only executed before each request that is handled by a function of
        that blueprint.
        """
        self.record_once(
            lambda s: s.app.before_request_funcs.setdefault(self.name, []).append(f)
        )
        return f

    def before_app_request(self, f):
        """Like :meth:`Flask.before_request`.  Such a function is executed
        before each request, even if outside of a blueprint.
        """
        self.record_once(
            lambda s: s.app.before_request_funcs.setdefault(None, []).append(f)
        )
        return f

    def before_app_first_request(self, f):
        """Like :meth:`Flask.before_first_request`.  Such a function is
        executed before the first request to the application.
        """
        self.record_once(lambda s: s.app.before_first_request_funcs.append(f))
        return f

    def after_request(self, f):
        """Like :meth:`Flask.after_request` but for a blueprint.  This function
        is only executed after each request that is handled by a function of
        that blueprint.
        """
        self.record_once(
            lambda s: s.app.after_request_funcs.setdefault(self.name, []).append(f)
        )
        return f

    def after_app_request(self, f):
        """Like :meth:`Flask.after_request` but for a blueprint.  Such a function
        is executed after each request, even if outside of the blueprint.
        """
        self.record_once(
            lambda s: s.app.after_request_funcs.setdefault(None, []).append(f)
        )
        return f

    def teardown_request(self, f):
        """Like :meth:`Flask.teardown_request` but for a blueprint.  This
        function is only executed when tearing down requests handled by a
        function of that blueprint.  Teardown request functions are executed
        when the request context is popped, even when no actual request was
        performed.
        """
        self.record_once(
            lambda s: s.app.teardown_request_funcs.setdefault(self.name, []).append(f)
        )
        return f

    def teardown_app_request(self, f):
        """Like :meth:`Flask.teardown_request` but for a blueprint.  Such a
        function is executed when tearing down each request, even if outside of
        the blueprint.
        """
        self.record_once(
            lambda s: s.app.teardown_request_funcs.setdefault(None, []).append(f)
        )
        return f

    def context_processor(self, f):
        """Like :meth:`Flask.context_processor` but for a blueprint.  This
        function is only executed for requests handled by a blueprint.
        """
        self.record_once(
            lambda s: s.app.template_context_processors.setdefault(
                self.name, []
            ).append(f)
        )
        return f

    def app_context_processor(self, f):
        """Like :meth:`Flask.context_processor` but for a blueprint.  Such a
        function is executed each request, even if outside of the blueprint.
        """
        self.record_once(
            lambda s: s.app.template_context_processors.setdefault(None, []).append(f)
        )
        return f

    def app_errorhandler(self, code):
        """Like :meth:`Flask.errorhandler` but for a blueprint.  This
        handler is used for all requests, even if outside of the blueprint.
        """

        def decorator(f):
            self.record_once(lambda s: s.app.errorhandler(code)(f))
            return f

        return decorator

    def url_value_preprocessor(self, f):
        """Registers a function as URL value preprocessor for this
        blueprint.  It's called before the view functions are called and
        can modify the url values provided.
        """
        self.record_once(
            lambda s: s.app.url_value_preprocessors.setdefault(self.name, []).append(f)
        )
        return f

    def url_defaults(self, f):
        """Callback function for URL defaults for this blueprint.  It's called
        with the endpoint and values and should update the values passed
        in place.
        """
        self.record_once(
            lambda s: s.app.url_default_functions.setdefault(self.name, []).append(f)
        )
        return f

    def app_url_value_preprocessor(self, f):
        """Same as :meth:`url_value_preprocessor` but application wide.
        """
        self.record_once(
            lambda s: s.app.url_value_preprocessors.setdefault(None, []).append(f)
        )
        return f

    def app_url_defaults(self, f):
        """Same as :meth:`url_defaults` but application wide.
        """
        self.record_once(
            lambda s: s.app.url_default_functions.setdefault(None, []).append(f)
        )
        return f

    def errorhandler(self, code_or_exception):
        """Registers an error handler that becomes active for this blueprint
        only.  Please be aware that routing does not happen local to a
        blueprint so an error handler for 404 usually is not handled by
        a blueprint unless it is caused inside a view function.  Another
        special case is the 500 internal server error which is always looked
        up from the application.

        Otherwise works as the :meth:`~flask.Flask.errorhandler` decorator
        of the :class:`~flask.Flask` object.
        """

        def decorator(f):
            self.record_once(
                lambda s: s.app._register_error_handler(self.name, code_or_exception, f)
            )
            return f

        return decorator

    def register_error_handler(self, code_or_exception, f):
        """Non-decorator version of the :meth:`errorhandler` error attach
        function, akin to the :meth:`~flask.Flask.register_error_handler`
        application-wide function of the :class:`~flask.Flask` object but
        for error handlers limited to this blueprint.

        .. versionadded:: 0.11
        """
        self.record_once(
            lambda s: s.app._register_error_handler(self.name, code_or_exception, f)
        )



import base64
import json
from xml.etree import cElementTree as eTree

from pos_model import TenderType
from typing import List

from fiscalpersistence import Tender


def convert_order_tenders_to_fiscal_tenders(order_picture):
    # type: (eTree) -> List[Tender]

    fiscal_tenders = []  # type: List[Tender]

    order_tenders = order_picture.findall('TenderHistory/Tender')

    for order_tender in order_tenders:
        fiscal_tender = Tender()
        fiscal_tender.tender_seq_id = order_tender.attrib["tenderId"]
        fiscal_tender.type = int(order_tender.attrib["tenderType"])
        fiscal_tender.amount = float(order_tender.attrib["tenderAmount"])
        fiscal_tender.change = float(order_tender.attrib['change']) if 'change' in order_tender.attrib else 0

        tender_details = order_tender.attrib['tenderDetail'] if 'tenderDetail' in order_tender.attrib else None
        if tender_details is not None:
            tender_details = json.loads(tender_details)

        if fiscal_tender.type in (TenderType.credit, TenderType.debit) and tender_details:
            fiscal_tender.auth_code = tender_details['AuthCode'] if 'AuthCode' in tender_details else ""
            fiscal_tender.bandeira = tender_details['Bandeira'] if 'Bandeira' in tender_details else ""
            fiscal_tender.cnpj_auth = tender_details['CNPJAuth'] if 'CNPJAuth' in tender_details else ""

            receipt_customer = tender_details['ReceiptCustomer'] if 'ReceiptCustomer' in tender_details else ""
            receipt_merchant = tender_details['ReceiptMerchant'] if 'ReceiptMerchant' in tender_details else ""
            if receipt_customer != "":
                receipt_customer = base64.b64decode(receipt_customer)
            if receipt_merchant != "":
                receipt_merchant = base64.b64decode(receipt_merchant)

            fiscal_tender.receipt_customer = receipt_customer
            fiscal_tender.receipt_merchant = receipt_merchant

        fiscal_tenders.append(fiscal_tender)

    return fiscal_tenders



import os
from logging import Logger

import cfgtools
import logging

from helper import import_pydevd, config_logger
from msgbus import MBEasyContext
from messagehandlerbuilder import ProductionMessageHandlerBuilder
from mbcontextmessagehandler import MbContextMessageBus, MbContextMessageHandler
from mwrepository import ProductRepository, ProductionRepository
from production.model import NoBoxesConfiguredException, NoViewsConfiguredException
from production.treebuilder import ViewConfiguration, BoxConfiguration, TreeConfig
from typing import List, Dict


def main():
    import_pydevd(os.environ["LOADERCFG"], 9131, False)
    config_logger(os.environ["LOADERCFG"], "Production", max_files=5, max_size=12582912, add_error_handler=True)
    config_logger(os.environ["LOADERCFG"], "ProductionPerformanceEvent", max_files=5, max_size=12582912)
    config_logger(os.environ["LOADERCFG"], "ProductionPerformanceCommand", max_files=5, max_size=12582912)

    base_logger = logging.getLogger("Production")
    debug_logger = build_logger(base_logger, "ProductionDebug", logging.DEBUG)
    info_logger = build_logger(base_logger, "ProductionInfo", logging.INFO)
    warning_logger = build_logger(base_logger, "ProductionWarning", logging.WARNING)
    error_logger = build_logger(base_logger, "ProductionError", logging.ERROR)

    loggers = {
        u"DEBUG": debug_logger,
        u"INFO": info_logger,
        u"WARNING": warning_logger,
        u"ERROR": error_logger
    }

    config = cfgtools.read(os.environ["LOADERCFG"])
    boxes_config = parse_box_configurations(config, "Production.Boxes")
    views_config = parse_view_configurations(config)
    branches_config = parse_box_configurations(config, "Production.Branches")

    if len(boxes_config) == 0:
        raise NoBoxesConfiguredException("No boxes configured")
    if len(views_config) == 0:
        raise NoViewsConfiguredException("No views configured")

    purge_interval = int(config.find_value("Production.PurgeInterval", "1"))
    orders_life_time = int(config.find_value("Production.OrdersLifeTime", "20"))
    manager_log_level = config.find_value("Production.Manager.LogLevel", "ERROR")
    audit_logger_prod_states = config.find_values("Production.Manager.AuditLogProdStates") or []
    kds_manager = config.find_value("Production.KdsManager", "NKDSCTRL")

    mb_context = MBEasyContext("NewProduction")
    mw_product_repository = ProductRepository(mb_context)
    mw_production_repository = ProductionRepository(mb_context, error_logger)

    mb_context.MB_EasyEvtSubscribe("ORDER_MODIFIED")
    message_bus = MbContextMessageBus(mb_context)

    message_handler_builder = ProductionMessageHandlerBuilder(
        message_bus,
        purge_interval,
        orders_life_time,
        mw_product_repository,
        mw_production_repository,
        manager_log_level,
        audit_logger_prod_states,
        boxes_config,
        branches_config,
        views_config,
        kds_manager,
        loggers
    )

    message_handler = MbContextMessageHandler(
        message_bus,
        "ProductionSystem",
        "Production",
        "",
        message_handler_builder)
    return message_handler.handle_events()


def parse_view_configurations(config):
    # type: (cfgtools.Group) -> List[ViewConfiguration]
    ret = []
    views_group = config.find_group("Production.Views")  # type: cfgtools.Group
    for group in views_group.groups:
        name = group.name
        if name is None:
            raise Exception("View without name")

        view_type = None
        log_level = "ERROR"
        extra_config = {}
        for key in group.keys:
            if key.name == "Type":
                view_type = key.value()
            elif key.name == "LogLevel":
                log_level = key.value()
            else:
                extra_config[key.name] = key.values if len(key.values) > 1 else key.value()

        if view_type is None:
            raise Exception("View without type: {}".format(name))

        ret.append(ViewConfiguration(name, view_type, log_level, extra_config))

    return ret


def parse_box_configurations(config, config_path):
    # type: (cfgtools.Group, str) -> List[TreeConfig]
    ret = []
    boxes_group = config.find_group(config_path)  # type: cfgtools.Group
    if not boxes_group:
        return ret

    for tree in boxes_group.groups:  # type: cfgtools.Group
        tree_boxes = []
        branch_name = tree.name
        for group in tree.groups:
            name = group.name.strip()
            if name is None:
                raise Exception("View without name")

            box_type = None
            log_level = "ERROR"
            extra_config = {}
            sons = []
            for key in group.keys:
                if key.name == "Type":
                    box_type = key.value().strip()
                elif key.name == "LogLevel":
                    log_level = key.value().strip()
                elif key.name == "Sons":
                    if len(key.values) > 0:
                        sons.extend([x.strip() for x in key.values])
                    else:
                        sons.append(key.value().strip())
                else:
                    extra_config[key.name] = key.values if len(key.values) > 1 else key.value()

            for inner_group in group.groups:
                extra_config[inner_group.name] = get_extra_config(inner_group)

            if box_type is None:
                raise Exception("View without type: {}".format(name))

            tree_boxes.append(BoxConfiguration(name, box_type, sons, log_level, extra_config))
        ret.append(TreeConfig(branch_name, tree_boxes))

    return ret


def get_extra_config(group):
    # type: (cfgtools.Group) -> Dict[str, str]
    inner_dict = {}
    for key in group.keys:
        inner_dict[key.name] = key.values if len(key.values) > 1 else key.value()
    for inner_group in group.groups:
        inner_dict[inner_group.name] = get_extra_config(inner_group)
    return inner_dict


def build_logger(base_logger, name, level):
    # type: (Logger, str, int) -> Logger
    logger = logging.getLogger(name)
    for handler in base_logger.handlers:
        logger.addHandler(handler)
    logger.setLevel(level)

    return logger



from ._MsgBusInfoMessage import MsgBusInfoMessage
from ._MsgBusScreenChanger import MsgBusScreenChanger
from ._MsgBusMessageBox import MsgBusMessageBox
from ._MsgBusListBox import MsgBusListBox

__all__ = [
    "MsgBusInfoMessage",
    "MsgBusScreenChanger",
    "MsgBusMessageBox",
    "MsgBusListBox"
]



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import absolute_import, division, print_function

from cryptography import utils
from cryptography.exceptions import (
    AlreadyFinalized, UnsupportedAlgorithm, _Reasons
)
from cryptography.hazmat.backends.interfaces import HMACBackend
from cryptography.hazmat.primitives import hashes


@utils.register_interface(hashes.HashContext)
class HMAC(object):
    def __init__(self, key, algorithm, backend, ctx=None):
        if not isinstance(backend, HMACBackend):
            raise UnsupportedAlgorithm(
                "Backend object does not implement HMACBackend.",
                _Reasons.BACKEND_MISSING_INTERFACE
            )

        if not isinstance(algorithm, hashes.HashAlgorithm):
            raise TypeError("Expected instance of hashes.HashAlgorithm.")
        self._algorithm = algorithm

        self._backend = backend
        self._key = key
        if ctx is None:
            self._ctx = self._backend.create_hmac_ctx(key, self.algorithm)
        else:
            self._ctx = ctx

    algorithm = utils.read_only_property("_algorithm")

    def update(self, data):
        if self._ctx is None:
            raise AlreadyFinalized("Context was already finalized.")
        utils._check_byteslike("data", data)
        self._ctx.update(data)

    def copy(self):
        if self._ctx is None:
            raise AlreadyFinalized("Context was already finalized.")
        return HMAC(
            self._key,
            self.algorithm,
            backend=self._backend,
            ctx=self._ctx.copy()
        )

    def finalize(self):
        if self._ctx is None:
            raise AlreadyFinalized("Context was already finalized.")
        digest = self._ctx.finalize()
        self._ctx = None
        return digest

    def verify(self, signature):
        utils._check_bytes("signature", signature)
        if self._ctx is None:
            raise AlreadyFinalized("Context was already finalized.")

        ctx, self._ctx = self._ctx, None
        ctx.verify(signature)




import sysactions

from typing import List, Dict  # noqa
from basereport import TableReport, ReportColumnDefinition, AlignTypes
from report import Part
from report.command import RepeatCommand, AlignCommand, FontCommand
from utils import format_line

from ._ReportBuilder import ReportBuilder
from ..dto import TableReportBodyDto, TableReportOrderDto, TableReportSaleLineDto


class TableReportBody(ReportBuilder):

    def __init__(self, table_report_body_dto, print_by, hide_unpriced_items, pre_account_with_discount, pos_id):
        # type: (TableReportBodyDto, str, bool, bool, str) -> None
        self.table_report_body_dto = table_report_body_dto
        self.print_by = print_by
        self.hide_unpriced_items = hide_unpriced_items
        self.pre_account_with_discount = pre_account_with_discount
        self.pos_id = pos_id

    def generate(self):
        # type: () -> List[TableReport]

        lines = []
        table_report_order_dto = self.table_report_body_dto.orders
        table_seats = self.table_report_body_dto.service_seats

        self \
            ._configure_font(lines) \
            ._generate_items_description(lines) \
            ._repeat_of_equals_symbol_line(lines) \
            ._generate_order(lines,
                             table_report_order_dto,
                             self.table_report_body_dto.tip,
                             self.table_report_body_dto.detailed_discounts,
                             self.table_report_body_dto.detailed_payments,
                             self.table_report_body_dto.detailed_tip_excluded,
                             self.table_report_body_dto.sub_total_amount,
                             self.table_report_body_dto.discount_amount,
                             self.table_report_body_dto.promotional_tenders,
                             table_seats) \
            ._repeat_of_equals_symbol_line(lines) \
            ._generate_sub_total_summary(lines,
                                         self.table_report_body_dto.sub_total_amount,
                                         self.table_report_body_dto.tip,
                                         self.table_report_body_dto.discount_amount,
                                         self.table_report_body_dto.promotional_tenders) \
            ._generate_total_summary(lines,
                                     self.table_report_body_dto.total_amount) \
            ._repeat_of_equals_symbol_line(lines) \
            ._generate_total_division_sugestion(lines,
                                                self.table_report_body_dto.total_amount,
                                                self.table_report_body_dto.service_seats) \
            ._generate_customer_doc_field(lines) \
            ._repeat_of_under_symbol_line(lines) \
            ._a_new_empty_line(lines) \
            ._generate_footer(lines)

        return lines

    def _generate_order(self, lines, table_report_order_dto, tip_amount, detailed_discounts, detailed_payments,
                        detailed_tip_excluded, sub_total_amount, total_discount, total_payment, seats):
        # type: (List, List[TableReportOrderDto], float, Dict, Dict, Dict, float, float, float, int) -> TableReportBody
        grouped_sale_lines = self._get_grouped_sale_lines(table_report_order_dto, seats)

        if self.print_by == "SEATS" and len(grouped_sale_lines) > 1:
            lines.extend(self._generate_orders_split_by_seats_from_sale_lines_with_tab_on_modifiers(
                    grouped_sale_lines, tip_amount, detailed_discounts, detailed_payments, detailed_tip_excluded,
                sub_total_amount, total_discount, total_payment))
        else:
            for order in table_report_order_dto:
                lines.extend(self._generate_order_without_split_seats(order))
            lines.extend([a_new_line()])

        return self

    @staticmethod
    def _get_parent_table_seat(sales_line_list, line_number):
        seat = [x.seat for x in sales_line_list if x.level == 0 and x.line_number == line_number]
        if len(seat) > 0:
            return seat[0]
        return 0

    def _get_table_seat(self, sales_line_list, sale_line):
        if sale_line.level == 0:
            seat = sale_line.seat
        else:
            seat = self._get_parent_table_seat(sales_line_list, sale_line.line_number)

        if seat == 0:
            seat = "$TABLE"

        return seat

    def _get_grouped_sale_lines(self, table_report_order_dto, seats):
        grouped_sale_lines = {}
        for i in range(seats):
            grouped_sale_lines[i+1] = []

        for order in table_report_order_dto:
            for sale_line in order.sales_line_list:
                if sale_line.item_type == "COUPON":
                    continue

                if sale_line.item_type == "COMBO" and sale_line.qty == 0:
                    continue

                seat = self._get_table_seat(order.sales_line_list, sale_line)

                if seat not in grouped_sale_lines:
                    grouped_sale_lines[seat] = [sale_line]
                else:
                    grouped_sale_lines[seat].append(sale_line)
        return grouped_sale_lines

    def _generate_order_without_split_seats(self, table_report_order_dto):
        # type: (TableReportOrderDto) -> List

        lines = []

        order_items_lines = self._build_product_lines(table_report_order_dto.sales_line_list)

        order_lines = TableReport(lines=order_items_lines, column_definitions=a_sub_total_summary_column_definition())
        lines.extend([order_lines])

        return lines

    def _generate_orders_split_by_seats_from_sale_lines_with_tab_on_modifiers(self,
                                                                              grouped_sale_lines,
                                                                              tip_amount,
                                                                              detailed_discounts,
                                                                              detailed_payments,
                                                                              detailed_tip_excluded,
                                                                              sub_total_amount,
                                                                              total_discount,
                                                                              total_payment):
        # type: (dict, float, Dict, Dict, Dict, float, float, float) -> List

        tip_excluded_amount = sum(detailed_tip_excluded.values())
        tip_percentage = self._get_percentage(tip_amount, sub_total_amount - tip_excluded_amount)
        seat_quantities = len(grouped_sale_lines.keys()) - (1 if "$TABLE" in grouped_sale_lines else 0)

        not_mapped_discount = total_discount - sum(detailed_discounts.values())
        not_mapped_payment = total_payment - sum(detailed_payments.values())

        table_tip_excluded = detailed_tip_excluded[0] if 0 in detailed_tip_excluded.keys() else 0.0
        table_discount = detailed_discounts[0] if 0 in detailed_discounts.keys() else 0.0
        table_payment = detailed_payments[0] if 0 in detailed_payments.keys() else 0.0
        table_sale_line = grouped_sale_lines["$TABLE"] if "$TABLE" in grouped_sale_lines else {}
        table_sub_total = 0.0 + sum([float(x.price) if x.price and x.qty != 0 else 0.0 for x in table_sale_line])
        table_tip = ((tip_percentage * (table_sub_total - table_tip_excluded)) / 100)
        table_per_seat = round(table_sub_total / seat_quantities, 2)
        table_discount_per_seat = table_discount / seat_quantities
        table_payment_per_seat = table_payment / seat_quantities
        table_tip_per_seat = table_tip / seat_quantities

        sub_total_seats_amount = sub_total_amount - table_sub_total

        lines = []
        for index, seat in enumerate(sorted(grouped_sale_lines, key=lambda g: str(g))):
            seat_lines = []
            add_to_report = True
            if index != 0:
                seat_lines.append(Part(commands=[RepeatCommand(38, "-")]))
                seat_lines.append(a_new_line())

            is_table = seat == "$TABLE"
            seat_descr = seat if is_table else "$SEAT_NUMBER {}".format(seat)
            seat_sub_total = sum([float(x.price) if x.price and x.qty != 0 else 0.0 for x in grouped_sale_lines[seat]])
            if is_table:
                seat_discount = 0
                seat_payment = 0
                seat_tip = 0
                seat_total = 0
            else:
                seat_tip_excluded = detailed_tip_excluded[seat] if seat in detailed_tip_excluded.keys() else 0.0
                seat_discount = not_mapped_discount * (seat_sub_total / sub_total_seats_amount)
                seat_discount += detailed_discounts[seat] if seat in detailed_discounts.keys() else 0.0
                seat_discount = round(seat_discount + table_discount_per_seat, 2)
                seat_payment = not_mapped_payment * (seat_sub_total / sub_total_seats_amount)
                seat_payment += detailed_payments[seat] if seat in detailed_payments.keys() else 0.0
                seat_payment = round(seat_payment + table_payment_per_seat, 2)
                seat_tip = (tip_percentage * (seat_sub_total - seat_tip_excluded)) / 100
                seat_tip = round(seat_tip + table_tip_per_seat, 2)
                seat_total = seat_sub_total + table_per_seat - seat_discount + seat_tip - seat_payment

            seat_title = [
                [seat_descr, ""]
            ]
            seat_lines.append(TableReport(lines=seat_title, column_definitions=a_total_summary_column_definitions()))
            seat_lines.append(a_new_line())

            seat_details = []
            has_details = False
            if table_per_seat > 0 and is_table:
                seat_details.append(["$DIVISION_FOR_SEATS", "#NUMBER({})".format(table_per_seat)])
            if table_per_seat > 0 and not is_table:
                seat_details.append(["$DIVISION_FROM_TABLE", "#NUMBER({})".format(table_per_seat)])
            if seat_discount > 0:
                seat_details.append(["$DISCOUNTS", "#NUMBER({})".format(-seat_discount)])
            if seat_payment > 0 and not is_table:
                seat_details.append(["$PROMOTIONAL_TENDERS", "#NUMBER({})".format(-seat_payment)])
            if seat_tip > 0:
                seat_details.append(["$TIP", "#NUMBER({})".format(seat_tip)])

            if len(seat_details) > 0:
                has_details = True

            if not is_table:
                seat_details.append(["$TOTAL", "#NUMBER({})".format(seat_total)])

            order_items = self._build_product_lines(grouped_sale_lines[seat])
            order_lines = TableReport(lines=order_items, column_definitions=a_sub_total_summary_column_definition())
            seat_lines.append(order_lines)
            seat_lines.append(a_new_line())
            if not has_details and len(order_lines.lines) == 0:
                add_to_report = False

            subtotal = [["$SUBTOTAL", "#NUMBER({})".format(seat_sub_total)]]
            seat_lines.append(TableReport(lines=subtotal, column_definitions=a_total_detail_summary_column_definition()))
            seat_lines.append(TableReport(lines=seat_details, column_definitions=a_total_detail_summary_column_definition()))

            if add_to_report:
                lines.extend(seat_lines)

        return lines

    @staticmethod
    def _get_percentage(amount, sub_total_amount):
        if amount > 0:
            percentage = round((amount * 100) / sub_total_amount, 2)
        else:
            percentage = 0
        return percentage

    def _build_product_lines(self, sales_line_list):
        # type: (List[TableReportSaleLineDto]) -> List

        if self.hide_unpriced_items:
            sales_line_list = self._remove_unpriced_items(sales_line_list)

        order_items_lines = []
        option_level = None
        for item in sales_line_list:
            if item.item_type in [None, "", "COUPON", "OPTION"]:
                option_level = item.level if item.item_type == "OPTION" else None
                continue

            if item.qty == 0 and (item.default_qty == 0 or item.qty < item.default_qty):
                option_level = None
                continue

            if item.level == 0:
                option_level = None

            item_service_period = "[{}]".format(item.service_period) if item.service_period else " "

            item_price = item.price or 0
            item_price = "#NUMBER({})".format(item_price) if item_price else ""

            indentation = " " * ((option_level or item.level) + 1)

            comment_description = ""
            if item.comment and item.comment.startswith("["):
                comment_description = "$" + item.comment.replace("[", "").replace("]", "").upper() + " "

            product_description = comment_description + item.product_name
            line_rjust = len(product_description) + len(indentation)
            description = product_description.rjust(line_rjust)
            qty_description = "{:.2f}".format(item.qty)
            order_items_lines.append([qty_description, description, item_service_period, item_price])

            if item.level <= option_level:
                option_level = None

        return order_items_lines

    @staticmethod
    def _remove_unpriced_items(sales_line_list):
        reduce_next_level = False
        removed_level = None
        lines_to_remove = []
        for line in sales_line_list:
            if line.price == 0:
                reduce_next_level = True
                removed_level = line.level
                lines_to_remove.append(line)
            else:
                if reduce_next_level and line.level > removed_level:
                    line.level -= 1
                elif line.level <= removed_level:
                    reduce_next_level = False
        for line in lines_to_remove:
            sales_line_list.remove(line)

        return sales_line_list

    def _generate_sub_total_summary(self, lines, sub_total_amount, tip_amount, discounts_amount, promotional_tenders):
        # type: (List, float, float, float, float) -> TableReportBody

        total_summary_column_definitions = a_total_summary_column_definitions()

        sub_total_lines = [["$SUBTOTAL", "#NUMBER({})".format(sub_total_amount)]]

        if discounts_amount > 0:
            sub_total_lines.append(["$DISCOUNTS", "#NUMBER({})".format(-discounts_amount)])
        if promotional_tenders > 0:
            sub_total_lines.append(["$PROMOTIONAL_TENDERS", "#NUMBER({})".format(-promotional_tenders)])
        if tip_amount > 0:
            sub_total_lines.append(["$SUGGESTED_TIP", "#NUMBER({})".format(tip_amount)])

        lines.extend(
            [TableReport(lines=sub_total_lines, column_definitions=total_summary_column_definitions), a_new_line()])
        return self

    def _generate_total_summary(self, lines, total_amount):
        # type: (List, float) -> TableReportBody

        total_summary_column_definitions = a_total_summary_column_definitions()

        sub_total_lines = [
            ["$TOTAL", "#NUMBER({})".format(total_amount)]
        ]
        lines.extend([TableReport(lines=sub_total_lines, column_definitions=total_summary_column_definitions)])
        return self

    def _generate_total_division_sugestion(self, lines, total_amount, service_seats):
        # type: (List, float, int) -> TableReportBody

        if service_seats > 1:
            division_amount = total_amount / service_seats

            lines.extend([
                Part("$EQUAL_DIVISION_FOR_ALL_SEATS", [AlignCommand(AlignCommand.Alignment.left)]), a_new_line()
            ])

            sub_total_lines = [
                ["$DIVISION_SUGESTION_FOR {}".format(service_seats), "#NUMBER({})".format(division_amount)]]

            lines.extend(
                [TableReport(lines=sub_total_lines, column_definitions=a_total_division_sugestion_definitions()),
                 a_new_line()
                 ])
        return self

    def _generate_customer_doc_field(self, lines):
        # type: (List) -> TableReportBody

        total_division_sugestion_definitions = a_customer_doc_field()

        sub_total_lines = [["$TYPE_YOUR_DOC_NUM"]]
        lines.extend(
            [TableReport(lines=sub_total_lines, column_definitions=total_division_sugestion_definitions),
             a_new_line()
             ])
        return self

    def _generate_footer(self, lines):
        # type: (List) -> TableReportBody
        lines.extend([
            Part("$THANK_YOU_FOR_YOUR_PREFERENCE", [AlignCommand(AlignCommand.Alignment.center)]), a_new_line(),
            Part("$SERVICE_TAX_IS_NOT_MANDATORY", [AlignCommand(AlignCommand.Alignment.center)]), a_new_line()
        ])

        if not self.pre_account_with_discount:
            model = sysactions.get_model(self.pos_id)
            message = sysactions.translate_message(model, "WARNING_REPORT_ITEMS_CONFERENCE")
            self._a_new_empty_line(lines)
            lines.extend(format_line('', message, columns=38, definition=a_information_footer()))

        self._a_new_empty_line(lines)
        return self

    def _generate_items_description(self, lines):
        # type: (List[TableReport]) -> TableReportBody

        items_description = TableReport(
            lines=[["$QTY", "$ITEM", "#CURRENCY_SYMBOL()"]],
            column_definitions=a_item_column_definition())

        lines.append(items_description)
        return self

    def _configure_font(self, lines):
        # type: (List[TableReport]) -> TableReportBody
        lines.append(Part(commands=[FontCommand(FontCommand.Font.A)]))

        return self

    def _repeat_of_equals_symbol_line(self, lines):
        # type: (List) -> TableReportBody
        repeat_symbol = [Part(commands=[RepeatCommand(38, "=")]), a_new_line()]
        lines.extend(repeat_symbol)

        return self

    def _repeat_of_under_symbol_line(self, lines):
        # type: (List) -> TableReportBody
        repeat_symbol = [Part(commands=[RepeatCommand(38, "_")]), a_new_line()]
        lines.extend(repeat_symbol)

        return self

    def _a_new_empty_line(self, lines):
        # type: (List) -> TableReportBody
        lines.extend([a_new_line()])

        return self


def a_new_line():
    return Part("\n")


def a_item_column_definition():
    return [
        ReportColumnDefinition(width=5,
                               before_text="",
                               after_text="",
                               after_fill_text="",
                               align=AlignTypes.LEFT),
        ReportColumnDefinition(width=25,
                               fill_with_char=" ",
                               before_text="  ",
                               align=AlignTypes.LEFT),
        ReportColumnDefinition(width=8,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.RIGHT)

    ]


def a_sub_total_summary_column_definition():
    return [
        ReportColumnDefinition(width=5,
                               before_text="",
                               after_text="",
                               after_fill_text="",
                               align=AlignTypes.CENTER),
        ReportColumnDefinition(width=22,
                               fill_with_char=" ",
                               before_text=" ",
                               align=AlignTypes.LEFT),
        ReportColumnDefinition(width=4,
                               fill_with_char=" ",
                               align=AlignTypes.RIGHT),
        ReportColumnDefinition(width=7,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.RIGHT)

    ]


def a_total_detail_summary_column_definition():
    return [
        ReportColumnDefinition(width=25,
                               fill_with_char=" ",
                               before_text=" ",
                               align=AlignTypes.RIGHT),
        ReportColumnDefinition(width=13,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.RIGHT)

    ]


def a_total_summary_column_definitions():
    return [
        ReportColumnDefinition(width=25,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.LEFT),
        ReportColumnDefinition(width=13,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.RIGHT)

    ]


def a_total_division_sugestion_definitions():
    return [
        ReportColumnDefinition(width=25,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.LEFT),
        ReportColumnDefinition(width=13,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.RIGHT)

    ]


def a_customer_doc_field():
    return [
        ReportColumnDefinition(width=25,
                               fill_with_char=" ",
                               before_text=" ",
                               align=AlignTypes.CENTER),
    ]


def a_information_footer():
    return [
        ReportColumnDefinition(width=38,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.CENTER),
    ]



# encoding: utf-8
import sys
import locale

from typing import List, Any

from systools import sys_log_exception

from basereport import TableReport, ReportColumnDefinition, SimpleReport, AlignTypes
from report import Presenter, Part
from report.command import NewLineCommand, AlignCommand, RepeatCommand, FontCommand
from helper import remove_accents

from reportsdata.salechallengereport.dto import SaleChallengeHeaderDto, SaleChallengeGeneratorDto
from salechallengereport.model import OperatorSaleItem, ProductSaleItem


class SaleChallengePresenter(Presenter):
    def __init__(self, columns):
        self.columns = columns

        self.header_title_column_definition = [
            ReportColumnDefinition(width=columns),
        ]
        self.header_column_definition = [
            ReportColumnDefinition(width=16, after_fill_text=': ', fill_with_char='.'),
            ReportColumnDefinition(width=columns - 16)
        ]

        self.body_operator_title_column_definition = [
            ReportColumnDefinition(width=32, align=AlignTypes.LEFT),
            ReportColumnDefinition(width=columns - 32, align=AlignTypes.CENTER)
        ]

        self.body_operator_item_column_definition = [
            ReportColumnDefinition(width=32),
            ReportColumnDefinition(
                    width=columns - 32, before_fill_text="[", after_fill_text="]", align=AlignTypes.RIGHT
            )
        ]

        self.body_product_title_column_definition = [
            ReportColumnDefinition(width=18, align=AlignTypes.LEFT),
            ReportColumnDefinition(width=columns - 18, before_fill_text=" ", align=AlignTypes.LEFT)
        ]

        self.body_product_item_column_definition = [
            ReportColumnDefinition(width=18),
            ReportColumnDefinition(width=6, before_fill_text="[", after_fill_text="]", align=AlignTypes.RIGHT),
            ReportColumnDefinition(width=14, before_fill_text=" R$", fill_with_char=" ", align=AlignTypes.RIGHT)
        ]

        try:
            locale.setlocale(locale.LC_ALL, "portuguese_brazil")
        except Exception as _:
            locale.setlocale(locale.LC_ALL, "pt_BR.UTF-8")

    def present(self, dto):
        # type: (SaleChallengeGeneratorDto) -> SimpleReport
        try:
            parts = [Part(commands=[FontCommand(FontCommand.Font.B)])]
            parts.extend(self._build_header(dto.header))
            parts.extend(self._build_items_title(dto.columns))
            parts.extend(self._build_items(dto.items, dto.columns))
            parts.append(Part(None, [NewLineCommand()]))
        except Exception as _:
            sys_log_exception("Error building Sale Challenge Report")
            raise

        return SimpleReport(parts, self.columns)

    def _build_header(self, header):
        parts = [
            Part(None, [RepeatCommand(self.columns, "="), NewLineCommand()]),
            self._get_title_report(),
            Part(None, [NewLineCommand()])
        ]
        parts.extend(self._get_sale_challenge_name(header))
        parts.extend(self._get_sale_challenge_period(header))
        parts.extend(self._get_sale_challenge_operator(header))
        parts.append(Part(None, [NewLineCommand()]))

        return parts

    def _build_items_title(self, columns):
        # type: (int) -> List[TableReport]

        parts = [
            Part(None, [RepeatCommand(self.columns, "="), NewLineCommand()]),
            self._get_items_title(columns),
            Part(None, [RepeatCommand(self.columns, "="), NewLineCommand()])
        ]

        return parts

    def _build_items(self, items, columns):
        # type: (Any[List[OperatorSaleItem], List[ProductSaleItem]], int) -> List[TableReport]

        rows = []
        two_columns = columns == 2
        column_definition = \
            self.body_operator_item_column_definition if two_columns else self.body_product_item_column_definition

        total_price = 0
        total_quantity = 0
        for item in items:
            total_quantity += item.quantity
            total_price += item.total_price if not two_columns else 0
            name = item.operator_name if two_columns else item.name
            name = remove_accents(name)
            quantity = self._str_int(item.quantity)
            price = self._str_float(item.total_price) if not two_columns else 0
            row = [[name, quantity]] if two_columns else [[name, quantity, price]]
            rows.append(TableReport(row, column_definition))

        rows.append(Part(None, [RepeatCommand(self.columns, "="), NewLineCommand()]))

        total_row = ["$TOTAL", self._str_int(total_quantity)]
        if not two_columns:
            total_row.append(self._str_float(total_price))
        rows.append(TableReport([total_row], column_definition))

        return rows

    def _format_line(self, title, text, definition):
        # type: (str, str, List[ReportColumnDefinition]) -> List[TableReport]

        if not text:
            return []

        lines = []
        current_string_start = 0
        current_string_end = definition[1].width if title else self.columns

        while len(text) > 0:
            if title and current_string_start == 0:
                lines.append(TableReport([[title, text[:current_string_end]]], definition))
            else:
                lines.append(TableReport([[text[current_string_start:current_string_end]]],
                                         [ReportColumnDefinition(width=self.columns)]))

            if current_string_end > len(text):
                break

            current_string_start = current_string_end
            current_string_end += self.columns

        return lines

    @staticmethod
    def _get_title_report():
        return Part("$CHALLENGE_REPORT", [AlignCommand(AlignCommand.Alignment.center)])

    def _get_sale_challenge_name(self, header):
        return self._format_line("$CHALLENGE", remove_accents(header.sale_challenge_name),
                                 self.header_column_definition)

    def _get_sale_challenge_period(self, header):
        # type: (SaleChallengeHeaderDto) -> List[TableReport]

        return self._format_line("$PERIOD", "{} - {}".format(header.start_date.strftime("%d/%m/%y"),
                                                             header.end_date.strftime("%d/%m/%y")),
                                 self.header_column_definition)

    def _get_items_title(self, columns):
        if columns == 2:
            return TableReport([["$NAME", "$QUANTITY_SHORT"]], self.body_operator_title_column_definition)
        return TableReport([["$NAME", "$QUANTITY_SHORT"]], self.body_product_title_column_definition)

    def _get_sale_challenge_operator(self, header):
        # type: (SaleChallengeHeaderDto) -> List[TableReport]

        if header.operator_name is None:
            return self._format_line("$OPERATOR", "$ALL_PRODUCTS", self.header_column_definition)
        return self._format_line("$OPERATOR", remove_accents(header.operator_name), self.header_column_definition)

    def _get_order_number(self, header):
        return self._format_line("$ORDER_DELIVERY_REPORT", header.order_number, self.header_column_definition)

    @staticmethod
    def _str_float(float_num):
        return locale.format("%.2f", float_num, True).replace(".", ",")

    @staticmethod
    def _str_int(float_num):
        return locale.format("%.0f", float_num, True)



import re

from production.model import Course
from typing import List, Tuple, Dict

from mw_helper import BaseRepository
from persistence import Connection

from production.repository import ProductRepository as ProductionProductRepository


class ProductRepository(BaseRepository, ProductionProductRepository):
    def __init__(self, mb_context):
        super(ProductRepository, self).__init__(mb_context)
        self.all_part_code_jit_lines = None

    def get_part_codes_of_jit_lines(self, jit_lines):
        # type: (List[str]) -> List[int]

        if len(jit_lines) == 0:
            return []

        part_codes = []
        all_products_jit_lines = self.get_all_part_code_jit_lines()
        for key, value in all_products_jit_lines.items():
            current_jit_lines = re.findall('(\\w+)', value)
            if any(x in jit_lines for x in current_jit_lines):
                part_codes.append(key)

        return part_codes

    def get_all_part_code_jit_lines(self):
        # type: () -> Dict[int, str]

        def func(conn):
            # type: (Connection) -> Dict[int, str]

            cursor = conn.select(get_part_codes_of_jit_lines_query)
            products_jit_lines = {}
            for row in cursor:
                products_jit_lines[int(row.get_entry("ProductCode"))] = row.get_entry("JITLines")

            return products_jit_lines

        if self.all_part_code_jit_lines is None:
            self.all_part_code_jit_lines = self.execute_with_connection(func)

        return self.all_part_code_jit_lines

    def get_combos_to_keep(self):
        def func(conn):
            # type: (Connection) -> List[int]
            cursor = conn.select(get_combos_to_keep_query)
            part_codes = []
            for row in cursor:
                part_codes.append(int(row.get_entry(0)))

            return part_codes

        return self.execute_with_connection(func)

    def get_master_product_map(self):
        def func(conn):
            # type: (Connection) -> Dict[int, Tuple[int, str]]
            cursor = conn.select(get_master_product_map_query)
            master_product_map = {}
            for row in cursor:
                part_code = int(row.get_entry(0))
                master_part_code = int(row.get_entry(1))
                master_product_name = row.get_entry(2).decode("utf-8")

                master_product_map[part_code] = (master_part_code, master_product_name)

            return master_product_map

        return self.execute_with_connection(func)

    def get_prep_times(self):
        def func(conn):
            # type: (Connection) -> Dict[int, int]
            cursor = conn.select(get_prep_time_query)
            prep_time_dict = {}
            for row in cursor:
                part_code = int(row.get_entry(0))
                prep_time = int(row.get_entry(1))

                prep_time_dict[part_code] = prep_time

            return prep_time_dict

        return self.execute_with_connection(func)

    def get_courses_products(self):
        def func(conn):
            # type: (Connection) -> Dict[int, int]
            cursor = conn.select(course_query)
            course_dict = {}
            for row in cursor:
                part_code = int(row.get_entry(0))
                course = row.get_entry(1)

                if course not in course_dict:
                    course_dict[course] = []
                course_dict[course].append(part_code)

            return course_dict

        return self.execute_with_connection(func)

    def get_not_show_on_kitchen(self):
        def func(conn):
            # type: (Connection) -> Dict[int, int]
            cursor = conn.select(not_show_on_kitchen_query)
            ret = []
            for row in cursor:
                ret.append(int(row.get_entry(0)))

            return ret

        return self.execute_with_connection(func)

    def get_no_jit_part_codes(self):
        def func(conn):
            # type: (Connection) -> Dict[int, int]
            cursor = conn.select(no_jit_query)
            ret = []
            for row in cursor:
                ret.append(int(row.get_entry(0)))

            return ret

        return self.execute_with_connection(func)

    def get_cfh_items(self):
        def func(conn):
            # type: (Connection) -> Dict[int, int]
            cursor = conn.select(cfh_items_query)
            ret = []
            for row in cursor:
                ret.append(int(row.get_entry(0)))

            return ret

        return self.execute_with_connection(func)

    def get_product_points(self):
        def func(conn):
            # type: (Connection) -> Dict[int, int]
            cursor = conn.select(item_points_query)
            ret = {}
            for row in cursor:
                ret[int(row.get_entry(0))] = int(row.get_entry(1))

            return ret

        return self.execute_with_connection(func)

    def get_all_courses(self):
        # type: () -> List[Course]

        def func(conn):
            # type: (Connection) -> List[Course]

            courses = []
            for row in conn.select(get_courses_query):
                course_id = int(row.get_entry("Id"))
                course_name = row.get_entry("Name")
                course_description = row.get_entry("Description")
                course_abbreviation = row.get_entry("Abbreviation")
                course_delay = int(row.get_entry("Delay"))
                courses.append(Course(course_id, course_name, course_description, course_abbreviation, course_delay))

            return courses

        return self.execute_with_connection(func)


get_part_codes_of_jit_lines_query = """
SELECT ProductCode, JITLines
FROM 
productdb.Production 
WHERE (Expiration IS NULL OR Expiration > DATETIME('now')) AND JITLInes IS NOT NULL
"""

get_combos_to_keep_query = """
SELECT ProductCode
FROM 
productdb.ProductCustomParams 
WHERE (CustomParamId = 'SHOW_ON_JIT')
"""

get_master_product_map_query = """
SELECT t.ProductCode, t.CustomParamValue, p.ProductName FROM 
(
SELECT ProductCode, CustomParamValue
FROM 
ProductCustomParams 
WHERE (CustomParamId = 'KDS_MASTER_PART_CODE')
) t INNER JOIN Product p ON t.CustomParamValue = p.ProductCode
"""

get_prep_time_query = """
SELECT ProductCode, CustomParamValue 
FROM ProductCustomParams 
WHERE CustomParamId = 'PrepTime' AND CustomParamValue IS NOT NULL
"""


course_query = """
select ProductCode, CustomParamValue from ProductCustomParams where CustomParamId = 'Course'
"""

not_show_on_kitchen_query = """
SELECT p.ProductCode
FROM Production p
JOIN ProductKernelParams pkp ON pkp.ProductCode = p.ProductCode
JOIN ProductType pt ON pt.TypeId = pkp.ProductType
WHERE ShowOnKitchen == 0 AND pt.TypeDescr <> 'TYPE_OPTION'
"""

no_jit_query = """
select ProductCode from Production where JITLines is NULL and ShowOnKitchen = 1
"""

cfh_items_query = """
 SELECT ProductCode FROM ProductTags
WHERE LOWER(Tag) = 'cfh=true'
UNION
SELECT ProductCode FROM ProductCustomParams
WHERE LOWER(CustomParamId) = 'cfh' AND LOWER(CustomParamValue) = 'true'
"""

item_points_query = """
SELECT ProductCode, CustomParamValue from ProductCustomParams where CustomParamId = 'ProductionPoints'
"""

get_courses_query = """
SELECT Id, Name, Description, Abbreviation, Delay FROM Courses
"""



from typing import List

from ._UndoCommand import UndoCommand


class ChangeProdStateUndoCommand(UndoCommand):
    def __init__(self, order_id, view, previous_prod_state, disabled_commands):
        # type: (str, str, str, List[UndoCommand]) -> None

        super(ChangeProdStateUndoCommand, self).__init__(order_id, view)
        self.previous_prod_state = previous_prod_state
        self.disabled_commands = disabled_commands

    def undo(self, order):
        order.prod_state = self.previous_prod_state

        for undo_command in self.disabled_commands:
            undo_command.enable()



from threading import Lock

from msgbus import MBEasyContext
from messagehandler import MessageHandlerBuilder, MessageHandler
from messageprocessor import MessageProcessorMessageHandler, DefaultMessageProcessorExecutorFactory
from messageprocessorutil import LoggingProcessorCallback

from dailygoals_services.interactor.sales_challenge import GetSalesChallengeInteractor, InsertSalesChallengeInteractor,\
    DeleteSalesChallengeInteractor, UpdateSalesChallengeInteractor
from dailygoals_services.interactor.update_goals import UpdateGoalsInteractor
from dailygoals_services.interactor.update_total_sold import UpdateTotalSoldInteractor, UpdateTotalSoldOrderInteractor,\
    UpdateTotalSoldTableInteractor
from dailygoals_services.service import DailyGoalsService
from externals.processor import UpdateGoalsProcessor, UpdateTotalSoldProcessor, UpdateTotalSoldOrderProcessor,\
    UpdateTotalSoldTableProcessor, GetSalesChallengeProcessor, InsertSalesChallengeProcessor,\
    DeleteSalesChallengeProcessor, UpdateSalesChallengeProcessor
from externals.repository.database import DailyGoalsRepository
from externals.util import LoggerUtil, LogManager
from externals.messages import ListenedMessages, ListenedEvents

from old_helper import read_swconfig

database_lock = Lock()
logger = LoggerUtil.get_logger_name(__name__)


class ComponentMessageHandlerBuilder(MessageHandlerBuilder):
    def __init__(self, mb_context, persistence_name, goals_filename):
        # type: (MBEasyContext, str, str, str) -> None

        super(ComponentMessageHandlerBuilder, self).__init__()

        self.mb_context = mb_context
        self.persistence_name = persistence_name
        self.goals_filename = goals_filename
        self.store_id = None

        self.log_managers = None
        self.subject_counter = {}
        self.message_handler = None

    def set_store_id(self, store_id):
        self.store_id = store_id

    def build_singletons(self):
        # type: () -> None

        self.store_id = read_swconfig(self.mb_context, "Store.Id")
        self.log_managers = LogManager("Subject Counter", self.subject_counter)
        self.log_managers.start()

        daily_goals_repository = DailyGoalsRepository(self.mb_context, self.persistence_name, database_lock)

        daily_goals_service = DailyGoalsService(daily_goals_repository, self.mb_context)
        daily_goals_service.configure(self.store_id, self.goals_filename)

        update_goals = ListenedMessages.TK_DAILY_GOALS_UPDATE_GOALS.value
        update_total_sold = ListenedMessages.TK_DAILY_GOALS_UPDATE_TOTAL_SOLD.value
        get_sales_challenge = ListenedMessages.TK_DAILY_GOALS_GET_SALES_CHALLENGE.value
        insert_sales_challenge = ListenedMessages.TK_DAILY_GOALS_INSERT_SALES_CHALLENGE.value
        delete_sales_challenge = ListenedMessages.TK_DAILY_GOALS_DELETE_SALES_CHALLENGE.value
        update_sales_challenge = ListenedMessages.TK_DAILY_GOALS_UPDATE_SALES_CHALLENGE.value

        update_goals_interactor = UpdateGoalsInteractor(daily_goals_repository, daily_goals_service)
        update_total_sold_interactor = UpdateTotalSoldInteractor(daily_goals_repository, daily_goals_service)
        get_sales_challenge_interactor = GetSalesChallengeInteractor(daily_goals_repository)
        insert_sales_challenge_interactor = InsertSalesChallengeInteractor(daily_goals_repository)
        delete_sales_challenge_interactor = DeleteSalesChallengeInteractor(daily_goals_repository)
        update_sales_challenge_interactor = UpdateSalesChallengeInteractor(daily_goals_repository)

        message_processors = {
            update_goals: UpdateGoalsProcessor(update_goals_interactor),
            update_total_sold: UpdateTotalSoldProcessor(update_total_sold_interactor),
            get_sales_challenge: GetSalesChallengeProcessor(get_sales_challenge_interactor),
            insert_sales_challenge: InsertSalesChallengeProcessor(insert_sales_challenge_interactor),
            delete_sales_challenge: DeleteSalesChallengeProcessor(delete_sales_challenge_interactor),
            update_sales_challenge: UpdateSalesChallengeProcessor(update_sales_challenge_interactor)
        }

        order_modified = ListenedEvents.ORDER_MODIFIED.value
        table_closed = ListenedEvents.TABLE_CLOSED.value

        update_total_sold_from_order = UpdateTotalSoldOrderInteractor(daily_goals_repository, daily_goals_service)
        update_total_sold_from_table = UpdateTotalSoldTableInteractor(daily_goals_repository, daily_goals_service)

        event_processors = {
            order_modified: UpdateTotalSoldOrderProcessor(update_total_sold_from_order),
            table_closed: UpdateTotalSoldTableProcessor(update_total_sold_from_table)
        }

        callbacks = [LoggingProcessorCallback(logger)]
        self.message_handler = MessageProcessorMessageHandler(
            event_processors,
            message_processors,
            DefaultMessageProcessorExecutorFactory(callbacks)
        )

    def build_message_handler(self):
        # type: () -> MessageHandler
        return self.message_handler

    def destroy_message_handler(self, message_handler):
        # type: (MessageHandlerBuilder) -> None

        return

    def destroy_singletons(self):
        # type: () -> None

        if self.log_managers:
            self.log_managers.stop()



import math

from abc import ABCMeta
from typing import List

from ._OrderChangerProductionBox import OrderChangerProductionBox
from production.model import ProductionOrder


class OrderSplitterBox(OrderChangerProductionBox):
    __metaclass__ = ABCMeta

    def __init__(self, name, sons, order_changer, split_type, split_value, logger=None):
        super(OrderSplitterBox, self).__init__(name, sons, logger)
        self.order_changer = order_changer
        self.split_type = split_type
        self.split_value = split_value

    def change_order(self, order):
        # type: (ProductionOrder) -> List[ProductionOrder]

        item_qty = [x for x in order.items if x.qty > 0]
        if len(item_qty) <= 1 or self.split_type is None:
            return [order]

        if self.split_type.upper() == 'PREP_TIME':
            orders = self.split_prep_time(order)
        elif self.split_type.upper() == 'SALE_LINES':
            if self.split_value is None:
                return [order]
            orders = self.split_sale_line(order, float(self.split_value))
        elif self.split_type.upper() == 'FIRE_HOLD':
            orders = self.split_fire_hold(order)
        else:
            orders = [order]

        return orders

    @staticmethod
    def split_prep_time(order):
        # type: (ProductionOrder) -> List[ProductionOrder]

        orders = []
        split_index = 0

        for item in order.items:
            for temp_order in orders:
                if len(temp_order.items) > 0 and item.prep_time == temp_order.items[0].prep_time:
                    temp_order.items.append(item)
                    break
            else:
                new_order = order.clone([item], split_index)
                orders.append(new_order)
                split_index += 1

        return orders

    def split_sale_line(self, order, limit):
        # type: (ProductionOrder, float) -> List[ProductionOrder]

        temp_items = [x for x in order.items if not self._has_fire_or_hold(x)]
        if len(temp_items) < limit:
            return [order]

        orders = []
        split_index = 0
        box_limit = math.ceil(len(temp_items) / limit)
        items_per_box = math.ceil(len(temp_items) / box_limit)
        for item in temp_items:
            if len(orders) <= split_index:
                new_order = order.clone([], split_index)
                orders.append(new_order)

            orders[split_index].items.append(item)
            if len(orders[split_index].items) >= items_per_box:
                split_index += 1

        orders[0].items.extend([x for x in order.items if self._has_fire_or_hold(x)])
        return orders

    def split_fire_hold(self, order):
        # type: (ProductionOrder) -> List[ProductionOrder]

        active_lines = [x for x in order.items if x.qty > 0]
        if len(active_lines) <= 1:
            return [order]

        orders = []
        main_order_items = []
        split_index = 1
        for item in order.items:
            if not self._has_fire_or_hold(item):
                main_order_items.append(item)
            elif item.qty > 0:
                new_order = order.clone([item], split_index)
                clear_tag_time = item.get_last_tag_time("clear_hold")
                if clear_tag_time is not None:
                    new_order.display_time = clear_tag_time.strftime("%Y-%m-%dT%H:%M:%S")
                orders.append(new_order)
                split_index += 1

        self.order_changer.change_order(order.order_id, self._mark_items_as_split)

        new_order = order
        if len(orders) > 0:
            new_order = order.clone()
            new_order.order_id = str(order.order_id) + "-0"
        new_order.items = main_order_items
        orders.append(new_order)

        return orders

    @staticmethod
    def _mark_items_as_split(root_order):
        # type: (ProductionOrder) -> bool

        need_to_save_order = False
        for item in root_order.items:
            if "fire" in item.properties and "had_fire" not in item.tags:
                item.add_tag("had_fire")
                need_to_save_order = True

            if "hold" in item.properties and "had_hold" not in item.tags:
                item.add_tag("had_hold")
                need_to_save_order = True

            if "hold" not in item.properties and "had_hold" in item.tags and "clear_hold" not in item.tags:
                item.add_tag("clear_hold")
                need_to_save_order = True

        return need_to_save_order

    @staticmethod
    def _has_fire_or_hold(item):
        has_fire = "fire" in item.properties or "had_fire" in item.tags
        has_hold = "hold" in item.properties or "had_hold" in item.tags

        return has_fire or has_hold



from abc import ABCMeta, abstractmethod

from typing import List  # noqa
from datetime import datetime  # noqa

from _Transfer import Transfer  # noqa


class TransferRepository(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_transfers(self, initial_date, end_date):
        # type: (datetime, datetime) -> List[Transfer]
        raise NotImplementedError()



from commons.report import Generator
from salechallengereport import SaleChallengeReport
from dto import SaleChallengeHeaderDto, SaleChallengeGeneratorDto


class SaleChallengeGenerator(Generator):
    def __init__(self, order_repository, product_repository, sale_challenge_repository, sale_challenge_id, operator_id):
        self.order_repository = order_repository
        self.product_repository = product_repository
        self.sale_challenge_repository = sale_challenge_repository
        self.sale_challenge_id = sale_challenge_id
        self.operator_id = operator_id

    def generate_data(self):
        sale_challenge_report = SaleChallengeReport(self.sale_challenge_repository, self.product_repository,
                                                    self.order_repository, self.operator_id)
        operator_name = sale_challenge_report.get_operator_name()
        sale_challenge = sale_challenge_report.get_sale_challenge(self.sale_challenge_id)
        sale_items = sale_challenge_report.get_sales_items(sale_challenge)
        header = SaleChallengeHeaderDto(sale_challenge, operator_name)
        columns = 3 if operator_name else 2
        return SaleChallengeGeneratorDto(header, sale_items, columns)



from typing import Optional, Any
from ._DataType import DataType


class Message(object):
    def __init__(self, token, data_type=DataType.empty, data=None, timeout=-1):
        # type: (int, int, Optional[Any], int) -> None
        self.token = token
        self.data_type = data_type
        self.data = data
        self.timeout = timeout
        self.imp_message = None  # type: Any



from datetime import datetime
from typing import Optional


class Order(object):
    def __init__(self, order_id, discount_amount, total_gross, applied_date, operator, authorizer, delivery):
        # type: (int, float, float, datetime, Optional[int], Optional[int], bool) -> None
        self.order_id = order_id
        self.discount_amount = discount_amount
        self.total_gross = total_gross
        self.applied_date = applied_date
        self.operator = operator
        self.authorizer = authorizer
        self.delivery = delivery



from mw_helper import ensure_iterable, convert_to_dict
from production.box._OrderChangerProductionBox import OrderChangerProductionBox
from production.model import ProdStates


class OrderStateFilterBox(OrderChangerProductionBox):
    def __init__(self, name, sons, allowed_order_states, excluded_order_states, logger=None):
        super(OrderStateFilterBox, self).__init__(name, sons, logger)
        self.allowed_order_states = convert_to_dict(ensure_iterable(allowed_order_states))
        self.excluded_order_states = convert_to_dict(ensure_iterable(excluded_order_states))

    def change_order(self, order):
        if self.should_filter_order(order):
            order.prod_state = ProdStates.INVALID

        return order

    def should_filter_order(self, order):
        if self.allowed_order_states and order.state not in self.allowed_order_states:
            return True
        if self.excluded_order_states and order.state in self.excluded_order_states:
            return True
        return False



from pyasn1 import error

class TagMap:
    def __init__(self, posMap={}, negMap={}, defType=None):
        self.__posMap = posMap.copy()
        self.__negMap = negMap.copy()
        self.__defType = defType
        
    def __contains__(self, tagSet):
        return tagSet in self.__posMap or \
               self.__defType is not None and tagSet not in self.__negMap

    def __getitem__(self, tagSet):
        if tagSet in self.__posMap:
            return self.__posMap[tagSet]
        elif tagSet in self.__negMap:
            raise error.PyAsn1Error('Key in negative map')
        elif self.__defType is not None:
            return self.__defType
        else:
            raise KeyError()

    def __repr__(self):
        s = self.__class__.__name__ + '('
        if self.__posMap:
            s = s + 'posMap=%r, ' % (self.__posMap,)
        if self.__negMap:
            s = s + 'negMap=%r, ' % (self.__negMap,)
        if self.__defType is not None:
            s = s + 'defType=%r' % (self.__defType,)
        return s + ')'

    def __str__(self):
        s = self.__class__.__name__ + ':\n'
        if self.__posMap:
            s = s + 'posMap:\n%s, ' % ',\n '.join([ x.prettyPrintType() for x in self.__posMap.values()])
        if self.__negMap:
            s = s + 'negMap:\n%s, ' % ',\n '.join([ x.prettyPrintType() for x in self.__negMap.values()])
        if self.__defType is not None:
            s = s + 'defType:\n%s, ' % self.__defType.prettyPrintType()
        return s

    def clone(self, parentType, tagMap, uniq=False):
        if self.__defType is not None and tagMap.getDef() is not None:
            raise error.PyAsn1Error('Duplicate default value at %s' % (self,))
        if tagMap.getDef() is not None:
            defType = tagMap.getDef()
        else:
            defType = self.__defType
            
        posMap = self.__posMap.copy()
        for k in tagMap.getPosMap():
            if uniq and k in posMap:
                raise error.PyAsn1Error('Duplicate positive key %s' % (k,))
            posMap[k] = parentType

        negMap = self.__negMap.copy()
        negMap.update(tagMap.getNegMap())
        
        return self.__class__(
            posMap, negMap, defType,
            )

    def getPosMap(self): return self.__posMap.copy()
    def getNegMap(self): return self.__negMap.copy()
    def getDef(self): return self.__defType



from pos_model import Tax
from old_helper import BaseRepository
from msgbus import MBEasyContext
from persistence import Connection


class NTaxCalcRepository(BaseRepository):
    def __init__(self, mbcontext):
        # type: (MBEasyContext) -> None
        super(NTaxCalcRepository, self).__init__(mbcontext)

    def load_all_taxes(self):
        # type: () -> List[Tax]
        def inner_func(conn):
            # type: (Connection) -> List[Tax]
            taxes = [Tax(x.get_entry(0), x.get_entry(1), float(x.get_entry(2)), x.get_entry(3), x.get_entry(4), x.get_entry(5)) for x in conn.select("SELECT Code, Name, Rate, FiscalIndex, TaxProcessor, Parameters from Tax Order by TaxOrder")]

            return taxes

        return self.execute_with_connection(inner_func, service_name="FiscalPersistence")

    def load_all_products_from_tax(self, tax_code):
        # type: (unicode) -> List[unicode]
        def inner_func(conn):
            # type: (Connection) -> Dict[unicode, unicode]
            products = {int(x.get_entry(0)): int(x.get_entry(0)) for x in conn.select("SELECT ProductCode from ProductTax where TaxCode = '{0}'".format(tax_code))}

            return products

        return self.execute_with_connection(inner_func, service_name="FiscalPersistence")



# encoding: utf-8
from old_helper import BaseRepository


class TableRepository(BaseRepository):

    def __init__(self, mbcontext, pos_id):
        super(TableRepository, self).__init__(mbcontext)
        self.pos_id = pos_id

    def get_opened_tables(self):
        def inner_func(conn):
            query = self._OpenedTablesQuery
            tables = []
            for x in conn.select(query):
                table_id = x.get_entry("TableId")

                if table_id not in tables:
                    tables.append(table_id)

            return tables

        return self.execute_with_connection(inner_func)

    _OpenedTablesQuery = """
                            SELECT ts.TableId FROM TableService ts
                            LEFT JOIN ServiceOrders so ON so.ServiceId = ts.ServiceId
                            WHERE ts.FinishedTS IS NULL
                            UNION
                            SELECT ts.TableId FROM TableService ts
                            LEFT JOIN ServiceOrders so ON so.ServiceId = ts.ServiceId
                            LEFT JOIN RestaurantTable rt ON rt.TableId = ts.TableId
                            WHERE ts.FinishedTS IS NOT NULL AND rt.Status = 7 AND ts.tableId not LIKE 'TAB-%'
                         """



from application.service import OrderService, BenefitService
from application.model import Configurations
from typing import Dict, List, Optional


class OnOrderReopenInteractor(object):

    def __init__(self, configs, order_service, benefit_service):
        # type: (Configurations, OrderService, BenefitService) -> None

        self.configs = configs
        self.order_service = order_service
        self.benefit_service = benefit_service

    def execute(self, pos_id):
        # type: (int) -> None

        order_picture = self.order_service.get_order_picture(pos_id)

        state = order_picture.get("state", "") if order_picture else ""
        if state != "IN_PROGRESS":
            return

        order = self.order_service.get_order(pos_id, order_picture=order_picture)
        order_picture = order.order_picture
        remove_manual = order.table_id is None

        applied_benefits = self.order_service.get_applied_benefits_from_order_picture(pos_id, order_picture)
        if not applied_benefits:
            return

        benefits_to_remove = self._get_benefits_to_remove(applied_benefits)
        if len(benefits_to_remove) > 0:
            self.benefit_service.remove_benefits(pos_id, benefits_to_remove, order_picture, delete_manual=remove_manual)

        self.benefit_service.unset_applied_benefit(pos_id, ["IN_PROGRESS"])

    def _get_benefits_to_remove(self, applied_benefits):
        # type: (List) -> List

        benefits_to_remove = []
        for applied_benefit in applied_benefits:
            benefit_key_to_remove = self._get_benefit_id_to_remove(applied_benefit)
            if benefit_key_to_remove:
                benefits_to_remove.append(benefit_key_to_remove)

        return benefits_to_remove

    @staticmethod
    def _get_benefit_id_to_remove(applied_benefit):
        # type: (Dict) -> Optional[int]

        benefit_applied = applied_benefit.get("benefit_applied")
        if not benefit_applied:
            return None

        benefit = applied_benefit.get("benefit", {})
        added_sale_lines = len(applied_benefit.get("added_sale_lines")) > 0

        must_remove_added_line = benefit.get("automatic", "") not in ("ON_SALE", "ON_TABLE_OPEN")
        if added_sale_lines and not must_remove_added_line:
            return None

        return applied_benefit.get("benefit_key")
